# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..lesions import FillLesions


def test_FillLesions_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bin_mask=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        cwf=dict(
            argstr="-cwf %f",
        ),
        debug=dict(
            argstr="-debug",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_dilation=dict(
            argstr="-dil %d",
        ),
        in_file=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        lesion_mask=dict(
            argstr="-l %s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        match=dict(
            argstr="-match %f",
        ),
        other=dict(
            argstr="-other",
        ),
        out_datatype=dict(
            argstr="-odt %s",
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_lesions_filled.nii.gz",
            position=3,
        ),
        search=dict(
            argstr="-search %f",
        ),
        size=dict(
            argstr="-size %d",
        ),
        smooth=dict(
            argstr="-smo %f",
        ),
        use_2d=dict(
            argstr="-2D",
        ),
        verbose=dict(
            argstr="-v",
        ),
    )
    inputs = FillLesions.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FillLesions_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = FillLesions.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..stats import UnaryStats


def test_UnaryStats_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        larger_voxel=dict(
            argstr="-t %f",
            position=-3,
        ),
        mask_file=dict(
            argstr="-m %s",
            extensions=None,
            position=-2,
        ),
        operation=dict(
            argstr="-%s",
            mandatory=True,
            position=4,
        ),
    )
    inputs = UnaryStats.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_UnaryStats_outputs():
    output_map = dict(
        output=dict(),
    )
    outputs = UnaryStats.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for join expansion
"""
import pytest

from .... import config
from ... import engine as pe
from ....interfaces import base as nib
from ....interfaces.utility import IdentityInterface, Function, Merge
from ....interfaces.base import traits, File


class PickFirstSpec(nib.TraitedSpec):
    in_files = traits.List(File(exists=True), argstr="%s", position=2, mandatory=True)


class PickFirstOutSpec(nib.TraitedSpec):
    output1 = File(exists=True)


class PickFirst(nib.BaseInterface):
    input_spec = PickFirstSpec
    output_spec = PickFirstOutSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output1"] = self.inputs.in_files[0]
        return outputs


class IncrementInputSpec(nib.TraitedSpec):
    input1 = nib.traits.Int(mandatory=True, desc="input")
    inc = nib.traits.Int(usedefault=True, default_value=1, desc="increment")


class IncrementOutputSpec(nib.TraitedSpec):
    output1 = nib.traits.Int(desc="ouput")


class IncrementInterface(nib.SimpleInterface):
    input_spec = IncrementInputSpec
    output_spec = IncrementOutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        self._results["output1"] = self.inputs.input1 + self.inputs.inc
        return runtime


_sums = []

_sum_operands = []


class SumInputSpec(nib.TraitedSpec):
    input1 = nib.traits.List(nib.traits.Int, mandatory=True, desc="input")


class SumOutputSpec(nib.TraitedSpec):
    output1 = nib.traits.Int(desc="ouput")
    operands = nib.traits.List(nib.traits.Int, desc="operands")


class SumInterface(nib.SimpleInterface):
    input_spec = SumInputSpec
    output_spec = SumOutputSpec

    def _run_interface(self, runtime):
        global _sum
        global _sum_operands
        runtime.returncode = 0
        self._results["operands"] = self.inputs.input1
        self._results["output1"] = sum(self.inputs.input1)
        _sum_operands.append(self.inputs.input1)
        _sums.append(sum(self.inputs.input1))
        return runtime


_set_len = None
"""The Set interface execution result."""


class SetInputSpec(nib.TraitedSpec):
    input1 = nib.traits.Set(nib.traits.Int, mandatory=True, desc="input")


class SetOutputSpec(nib.TraitedSpec):
    output1 = nib.traits.Int(desc="ouput")


class SetInterface(nib.BaseInterface):
    input_spec = SetInputSpec
    output_spec = SetOutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        return runtime

    def _list_outputs(self):
        global _set_len
        outputs = self._outputs().get()
        _set_len = outputs["output1"] = len(self.inputs.input1)
        return outputs


_products = []
"""The Products interface execution results."""


class ProductInputSpec(nib.TraitedSpec):
    input1 = nib.traits.Int(mandatory=True, desc="input1")
    input2 = nib.traits.Int(mandatory=True, desc="input2")


class ProductOutputSpec(nib.TraitedSpec):
    output1 = nib.traits.Int(mandatory=True, desc="output")


class ProductInterface(nib.BaseInterface):
    input_spec = ProductInputSpec
    output_spec = ProductOutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        return runtime

    def _list_outputs(self):
        global _products
        outputs = self._outputs().get()
        outputs["output1"] = self.inputs.input1 * self.inputs.input2
        _products.append(outputs["output1"])
        return outputs


@pytest.mark.parametrize("needed_outputs", ["true", "false"])
def test_join_expansion(tmpdir, needed_outputs):
    global _sums
    global _sum_operands
    global _products
    tmpdir.chdir()

    # Clean up, just in case some other test modified them
    _products = []
    _sum_operands = []
    _sums = []

    prev_state = config.get("execution", "remove_unnecessary_outputs")
    config.set("execution", "remove_unnecessary_outputs", needed_outputs)
    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2])]
    # a pre-join node in the iterated path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    # another pre-join node in the iterated path
    pre_join2 = pe.Node(IncrementInterface(), name="pre_join2")
    # the join node
    join = pe.JoinNode(
        SumInterface(), joinsource="inputspec", joinfield="input1", name="join"
    )
    # an uniterated post-join node
    post_join1 = pe.Node(IncrementInterface(), name="post_join1")
    # a post-join node in the iterated path
    post_join2 = pe.Node(ProductInterface(), name="post_join2")

    wf.connect(
        [
            (inputspec, pre_join1, [("n", "input1")]),
            (pre_join1, pre_join2, [("output1", "input1")]),
            (pre_join1, post_join2, [("output1", "input2")]),
            (pre_join2, join, [("output1", "input1")]),
            (join, post_join1, [("output1", "input1")]),
            (join, post_join2, [("output1", "input1")]),
        ]
    )
    result = wf.run()

    # the two expanded pre-join predecessor nodes feed into one join node
    joins = [node for node in result.nodes() if node.name == "join"]
    assert len(joins) == 1, "The number of join result nodes is incorrect."
    # the expanded graph contains 2 * 2 = 4 iteration pre-join nodes, 1 join
    # node, 1 non-iterated post-join node and 2 * 1 iteration post-join nodes.
    # Nipype factors away the IdentityInterface.
    assert len(result.nodes()) == 8, "The number of expanded nodes is incorrect."

    # the join Sum result is (1 + 1 + 1) + (2 + 1 + 1)
    assert len(_sums) == 1, "The number of join outputs is incorrect"
    assert _sums[0] == 7, "The join Sum output value is incorrect: %s." % _sums[0]
    # the join input preserves the iterables input order
    assert _sum_operands[0] == [3, 4], (
        "The join Sum input is incorrect: %s." % _sum_operands[0]
    )
    # there are two iterations of the post-join node in the iterable path
    assert len(_products) == 2, "The number of iterated post-join outputs is incorrect"
    config.set("execution", "remove_unnecessary_outputs", prev_state)


def test_node_joinsource(tmpdir):
    """Test setting the joinsource to a Node."""
    tmpdir.chdir()

    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2])]
    # the join node
    join = pe.JoinNode(
        SetInterface(), joinsource=inputspec, joinfield="input1", name="join"
    )

    # the joinsource is the inputspec name
    assert (
        join.joinsource == inputspec.name
    ), "The joinsource is not set to the node name."


def test_set_join_node(tmpdir):
    """Test collecting join inputs to a set."""
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2, 1, 3, 2])]
    # a pre-join node in the iterated path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "input1")
    # the set join node
    join = pe.JoinNode(
        SetInterface(), joinsource="inputspec", joinfield="input1", name="join"
    )
    wf.connect(pre_join1, "output1", join, "input1")

    wf.run()

    # the join length is the number of unique inputs
    assert _set_len == 3, "The join Set output value is incorrect: %s." % _set_len


def test_unique_join_node(tmpdir):
    """Test join with the ``unique`` flag set to True."""
    global _sum_operands
    _sum_operands = []
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [3, 1, 2, 1, 3])]
    # a pre-join node in the iterated path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "input1")
    # the set join node
    join = pe.JoinNode(
        SumInterface(),
        joinsource="inputspec",
        joinfield="input1",
        unique=True,
        name="join",
    )
    wf.connect(pre_join1, "output1", join, "input1")

    wf.run()

    assert _sum_operands[0] == [4, 2, 3], (
        "The unique join output value is incorrect: %s." % _sum_operands[0]
    )


def test_multiple_join_nodes(tmpdir):
    """Test two join nodes, one downstream of the other."""
    global _products
    _products = []
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2, 3])]
    # a pre-join node in the iterated path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "input1")
    # the first join node
    join1 = pe.JoinNode(
        IdentityInterface(fields=["vector"]),
        joinsource="inputspec",
        joinfield="vector",
        name="join1",
    )
    wf.connect(pre_join1, "output1", join1, "vector")
    # an uniterated post-join node
    post_join1 = pe.Node(SumInterface(), name="post_join1")
    wf.connect(join1, "vector", post_join1, "input1")
    # the downstream join node connected to both an upstream join
    # path output and a separate input in the iterated path
    join2 = pe.JoinNode(
        IdentityInterface(fields=["vector", "scalar"]),
        joinsource="inputspec",
        joinfield="vector",
        name="join2",
    )
    wf.connect(pre_join1, "output1", join2, "vector")
    wf.connect(post_join1, "output1", join2, "scalar")
    # a second post-join node
    post_join2 = pe.Node(SumInterface(), name="post_join2")
    wf.connect(join2, "vector", post_join2, "input1")
    # a third post-join node
    post_join3 = pe.Node(ProductInterface(), name="post_join3")
    wf.connect(post_join2, "output1", post_join3, "input1")
    wf.connect(join2, "scalar", post_join3, "input2")

    result = wf.run()

    # The expanded graph contains one pre_join1 replicate per inputspec
    # replicate and one of each remaining node = 3 + 5 = 8 nodes.
    # The replicated inputspec nodes are factored out of the expansion.
    assert len(result.nodes()) == 8, "The number of expanded nodes is incorrect."
    # The outputs are:
    # pre_join1: [2, 3, 4]
    # post_join1: 9
    # join2: [2, 3, 4] and 9
    # post_join2: 9
    # post_join3: 9 * 9 = 81
    assert _products == [81], "The post-join product is incorrect"


def test_identity_join_node(tmpdir):
    """Test an IdentityInterface join."""
    global _sum_operands
    _sum_operands = []
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2, 3])]
    # a pre-join node in the iterated path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "input1")
    # the IdentityInterface join node
    join = pe.JoinNode(
        IdentityInterface(fields=["vector"]),
        joinsource="inputspec",
        joinfield="vector",
        name="join",
    )
    wf.connect(pre_join1, "output1", join, "vector")
    # an uniterated post-join node
    post_join1 = pe.Node(SumInterface(), name="post_join1")
    wf.connect(join, "vector", post_join1, "input1")

    result = wf.run()

    # the expanded graph contains 1 * 3 iteration pre-join nodes, 1 join
    # node and 1 post-join node. Nipype factors away the iterable input
    # IdentityInterface but keeps the join IdentityInterface.
    assert len(result.nodes()) == 5, "The number of expanded nodes is incorrect."
    assert _sum_operands[0] == [2, 3, 4], (
        "The join Sum input is incorrect: %s." % _sum_operands[0]
    )


def test_multifield_join_node(tmpdir):
    """Test join on several fields."""
    global _products
    _products = []
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["m", "n"]), name="inputspec")
    inputspec.iterables = [("m", [1, 2]), ("n", [3, 4])]
    # two pre-join nodes in a parallel iterated path
    inc1 = pe.Node(IncrementInterface(), name="inc1")
    wf.connect(inputspec, "m", inc1, "input1")
    inc2 = pe.Node(IncrementInterface(), name="inc2")
    wf.connect(inputspec, "n", inc2, "input1")
    # the join node
    join = pe.JoinNode(
        IdentityInterface(fields=["vector1", "vector2"]),
        joinsource="inputspec",
        name="join",
    )
    wf.connect(inc1, "output1", join, "vector1")
    wf.connect(inc2, "output1", join, "vector2")
    # a post-join node
    prod = pe.MapNode(ProductInterface(), name="prod", iterfield=["input1", "input2"])
    wf.connect(join, "vector1", prod, "input1")
    wf.connect(join, "vector2", prod, "input2")

    result = wf.run()

    # the iterables are expanded as the cartesian product of the iterables values.
    # thus, the expanded graph contains 2 * (2 * 2) iteration pre-join nodes, 1 join
    # node and 1 post-join node.
    assert len(result.nodes()) == 10, "The number of expanded nodes is incorrect."
    # the product inputs are [2, 4], [2, 5], [3, 4], [3, 5]
    assert set(_products) == set([8, 10, 12, 15]), (
        "The post-join products is incorrect: %s." % _products
    )


def test_synchronize_join_node(tmpdir):
    """Test join on an input node which has the ``synchronize`` flag set to True."""
    global _products
    _products = []
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["m", "n"]), name="inputspec")
    inputspec.iterables = [("m", [1, 2]), ("n", [3, 4])]
    inputspec.synchronize = True
    # two pre-join nodes in a parallel iterated path
    inc1 = pe.Node(IncrementInterface(), name="inc1")
    wf.connect(inputspec, "m", inc1, "input1")
    inc2 = pe.Node(IncrementInterface(), name="inc2")
    wf.connect(inputspec, "n", inc2, "input1")
    # the join node
    join = pe.JoinNode(
        IdentityInterface(fields=["vector1", "vector2"]),
        joinsource="inputspec",
        name="join",
    )
    wf.connect(inc1, "output1", join, "vector1")
    wf.connect(inc2, "output1", join, "vector2")
    # a post-join node
    prod = pe.MapNode(ProductInterface(), name="prod", iterfield=["input1", "input2"])
    wf.connect(join, "vector1", prod, "input1")
    wf.connect(join, "vector2", prod, "input2")

    result = wf.run()

    # there are 3 iterables expansions.
    # thus, the expanded graph contains 2 * 2 iteration pre-join nodes, 1 join
    # node and 1 post-join node.
    assert len(result.nodes()) == 6, "The number of expanded nodes is incorrect."
    # the product inputs are [2, 3] and [4, 5]
    assert _products == [8, 15], "The post-join products is incorrect: %s." % _products


def test_itersource_join_source_node(tmpdir):
    """Test join on an input node which has an ``itersource``."""
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2])]
    # an intermediate node in the first iteration path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "input1")
    # an iterable pre-join node with an itersource
    pre_join2 = pe.Node(ProductInterface(), name="pre_join2")
    pre_join2.itersource = ("inputspec", "n")
    pre_join2.iterables = ("input1", {1: [3, 4], 2: [5, 6]})
    wf.connect(pre_join1, "output1", pre_join2, "input2")
    # an intermediate node in the second iteration path
    pre_join3 = pe.Node(IncrementInterface(), name="pre_join3")
    wf.connect(pre_join2, "output1", pre_join3, "input1")
    # the join node
    join = pe.JoinNode(
        IdentityInterface(fields=["vector"]),
        joinsource="pre_join2",
        joinfield="vector",
        name="join",
    )
    wf.connect(pre_join3, "output1", join, "vector")
    # a join successor node
    post_join1 = pe.Node(SumInterface(), name="post_join1")
    wf.connect(join, "vector", post_join1, "input1")

    result = wf.run()

    # the expanded graph contains
    # 1 pre_join1 replicate for each inputspec iteration,
    # 2 pre_join2 replicates for each inputspec iteration,
    # 1 pre_join3 for each pre_join2 iteration,
    # 1 join replicate for each inputspec iteration and
    # 1 post_join1 replicate for each join replicate =
    # 2 + (2 * 2) + 4 + 2 + 2 = 14 expansion graph nodes.
    # Nipype factors away the iterable input
    # IdentityInterface but keeps the join IdentityInterface.
    assert len(result.nodes()) == 14, "The number of expanded nodes is incorrect."
    # The first join inputs are:
    # 1 + (3 * 2) and 1 + (4 * 2)
    # The second join inputs are:
    # 1 + (5 * 3) and 1 + (6 * 3)
    # the post-join nodes execution order is indeterminate;
    # therefore, compare the lists item-wise.
    assert [16, 19] in _sum_operands, (
        "The join Sum input is incorrect: %s." % _sum_operands
    )
    assert [7, 9] in _sum_operands, (
        "The join Sum input is incorrect: %s." % _sum_operands
    )


def test_itersource_two_join_nodes(tmpdir):
    """Test join with a midstream ``itersource`` and an upstream
    iterable."""
    tmpdir.chdir()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [("n", [1, 2])]
    # an intermediate node in the first iteration path
    pre_join1 = pe.Node(IncrementInterface(), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "input1")
    # an iterable pre-join node with an itersource
    pre_join2 = pe.Node(ProductInterface(), name="pre_join2")
    pre_join2.itersource = ("inputspec", "n")
    pre_join2.iterables = ("input1", {1: [3, 4], 2: [5, 6]})
    wf.connect(pre_join1, "output1", pre_join2, "input2")
    # an intermediate node in the second iteration path
    pre_join3 = pe.Node(IncrementInterface(), name="pre_join3")
    wf.connect(pre_join2, "output1", pre_join3, "input1")
    # the first join node
    join1 = pe.JoinNode(
        IdentityInterface(fields=["vector"]),
        joinsource="pre_join2",
        joinfield="vector",
        name="join1",
    )
    wf.connect(pre_join3, "output1", join1, "vector")
    # a join successor node
    post_join1 = pe.Node(SumInterface(), name="post_join1")
    wf.connect(join1, "vector", post_join1, "input1")
    # a summary join node
    join2 = pe.JoinNode(
        IdentityInterface(fields=["vector"]),
        joinsource="inputspec",
        joinfield="vector",
        name="join2",
    )
    wf.connect(post_join1, "output1", join2, "vector")

    result = wf.run()

    # the expanded graph contains the 14 test_itersource_join_source_node
    # nodes plus the summary join node.
    assert len(result.nodes()) == 15, "The number of expanded nodes is incorrect."


def test_set_join_node_file_input(tmpdir):
    """Test collecting join inputs to a set."""
    tmpdir.chdir()
    open("test.nii", "w+").close()
    open("test2.nii", "w+").close()

    # Make the workflow.
    wf = pe.Workflow(name="test")
    # the iterated input node
    inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
    inputspec.iterables = [
        ("n", [tmpdir.join("test.nii").strpath, tmpdir.join("test2.nii").strpath])
    ]
    # a pre-join node in the iterated path
    pre_join1 = pe.Node(IdentityInterface(fields=["n"]), name="pre_join1")
    wf.connect(inputspec, "n", pre_join1, "n")
    # the set join node
    join = pe.JoinNode(
        PickFirst(), joinsource="inputspec", joinfield="in_files", name="join"
    )
    wf.connect(pre_join1, "n", join, "in_files")

    wf.run()


def test_nested_workflow_join(tmpdir):
    """Test collecting join inputs within a nested workflow"""
    tmpdir.chdir()

    # Make the nested workflow
    def nested_wf(i, name="smallwf"):
        # iterables with list of nums
        inputspec = pe.Node(IdentityInterface(fields=["n"]), name="inputspec")
        inputspec.iterables = [("n", i)]
        # increment each iterable before joining
        pre_join = pe.Node(IncrementInterface(), name="pre_join")
        # rejoin nums into list
        join = pe.JoinNode(
            IdentityInterface(fields=["n"]),
            joinsource="inputspec",
            joinfield="n",
            name="join",
        )
        # define and connect nested workflow
        wf = pe.Workflow(name="wf_%d" % i[0])
        wf.connect(inputspec, "n", pre_join, "input1")
        wf.connect(pre_join, "output1", join, "n")
        return wf

    # master wf
    meta_wf = pe.Workflow(name="meta", base_dir=".")
    # add each mini-workflow to master
    for i in [[1, 3], [2, 4]]:
        mini_wf = nested_wf(i)
        meta_wf.add_nodes([mini_wf])

    result = meta_wf.run()

    # there should be six nodes in total
    assert len(result.nodes()) == 6, "The number of expanded nodes is incorrect."


def test_name_prefix_join(tmpdir):
    tmpdir.chdir()

    def sq(x):
        return x ** 2

    wf = pe.Workflow("wf", base_dir=tmpdir.strpath)
    square = pe.Node(Function(function=sq), name="square")
    square.iterables = [("x", [1, 2])]
    square_join = pe.JoinNode(
        Merge(1, ravel_inputs=True),
        name="square_join",
        joinsource="square",
        joinfield=["in1"],
    )
    wf.connect(square, "out", square_join, "in1")
    wf.run()


def test_join_nestediters(tmpdir):
    tmpdir.chdir()

    def exponent(x, p):
        return x ** p

    wf = pe.Workflow("wf", base_dir=tmpdir.strpath)

    xs = pe.Node(IdentityInterface(["x"]), iterables=[("x", [1, 2])], name="xs")
    ps = pe.Node(IdentityInterface(["p"]), iterables=[("p", [3, 4])], name="ps")
    exp = pe.Node(Function(function=exponent), name="exp")
    exp_joinx = pe.JoinNode(
        Merge(1, ravel_inputs=True),
        name="exp_joinx",
        joinsource="xs",
        joinfield=["in1"],
    )
    exp_joinp = pe.JoinNode(
        Merge(1, ravel_inputs=True),
        name="exp_joinp",
        joinsource="ps",
        joinfield=["in1"],
    )
    wf.connect(
        [
            (xs, exp, [("x", "x")]),
            (ps, exp, [("p", "p")]),
            (exp, exp_joinx, [("out", "in1")]),
            (exp_joinx, exp_joinp, [("out", "in1")]),
        ]
    )

    wf.run()



"""Image manipulation utilities (mostly, NiBabel manipulations)."""
import nibabel as nb


def copy_header(header_file, in_file, keep_dtype=True):
    """Copy header from a reference image onto another image."""
    hdr_img = nb.load(header_file)
    out_img = nb.load(in_file, mmap=False)
    hdr = hdr_img.header.copy()
    if keep_dtype:
        hdr.set_data_dtype(out_img.get_data_dtype())

    new_img = out_img.__class__(out_img.dataobj, None, hdr)
    if not keep_dtype:
        new_img.set_data_dtype(hdr_img.get_data_dtype())

    new_img.to_filename(in_file)
    return in_file



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The QT1 module of niftyfit, which wraps the Multi-Echo T1 fitting methods
in NiftyFit.
"""

from ..base import TraitedSpec, File, traits, CommandLineInputSpec
from .base import NiftyFitCommand
from ..niftyreg.base import get_custom_path


class FitQt1InputSpec(CommandLineInputSpec):
    """Input Spec for FitQt1."""

    desc = "Filename of the 4D Multi-Echo T1 source image."
    source_file = File(
        position=1, exists=True, desc=desc, argstr="-source %s", mandatory=True
    )

    # Output options:
    t1map_file = File(
        name_source=["source_file"],
        name_template="%s_t1map.nii.gz",
        argstr="-t1map %s",
        desc="Filename of the estimated output T1 map (in ms).",
    )
    m0map_file = File(
        name_source=["source_file"],
        name_template="%s_m0map.nii.gz",
        argstr="-m0map %s",
        desc="Filename of the estimated input M0 map.",
    )
    desc = "Filename of the estimated output multi-parameter map."
    mcmap_file = File(
        name_source=["source_file"],
        name_template="%s_mcmap.nii.gz",
        argstr="-mcmap %s",
        desc=desc,
    )
    comp_file = File(
        name_source=["source_file"],
        name_template="%s_comp.nii.gz",
        argstr="-comp %s",
        desc="Filename of the estimated multi-component T1 map.",
    )
    desc = "Filename of the error map (symmetric matrix, [Diag,OffDiag])."
    error_file = File(
        name_source=["source_file"],
        name_template="%s_error.nii.gz",
        argstr="-error %s",
        desc=desc,
    )
    syn_file = File(
        name_source=["source_file"],
        name_template="%s_syn.nii.gz",
        argstr="-syn %s",
        desc="Filename of the synthetic ASL data.",
    )
    res_file = File(
        name_source=["source_file"],
        name_template="%s_res.nii.gz",
        argstr="-res %s",
        desc="Filename of the model fit residuals",
    )

    # Other options:
    mask = File(
        position=2, exists=True, desc="Filename of image mask.", argstr="-mask %s"
    )
    prior = File(
        position=3, exists=True, desc="Filename of parameter prior.", argstr="-prior %s"
    )
    te_value = traits.Float(desc="TE Echo Time [0ms!].", argstr="-TE %f", position=4)
    tr_value = traits.Float(
        desc="TR Repetition Time [10s!].", argstr="-TR %f", position=5
    )
    desc = "Number of components to fit [1] (currently IR/SR only)"
    # set position to be ahead of TIs
    nb_comp = traits.Int(desc=desc, position=6, argstr="-nc %d")
    desc = "Set LM parameters (initial value, decrease rate) [100,1.2]."
    lm_val = traits.Tuple(
        traits.Float, traits.Float, desc=desc, argstr="-lm %f %f", position=7
    )
    desc = "Use Gauss-Newton algorithm [Levenberg-Marquardt]."
    gn_flag = traits.Bool(desc=desc, argstr="-gn", position=8)
    slice_no = traits.Int(
        desc="Fit to single slice number.", argstr="-slice %d", position=9
    )
    voxel = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        desc="Fit to single voxel only.",
        argstr="-voxel %d %d %d",
        position=10,
    )
    maxit = traits.Int(desc="NLSQR iterations [100].", argstr="-maxit %d", position=11)

    # IR options:
    sr_flag = traits.Bool(
        desc="Saturation Recovery fitting [default].", argstr="-SR", position=12
    )
    ir_flag = traits.Bool(
        desc="Inversion Recovery fitting [default].", argstr="-IR", position=13
    )
    tis = traits.List(
        traits.Float,
        position=14,
        desc="Inversion times for T1 data [1s,2s,5s].",
        argstr="-TIs %s",
        sep=" ",
    )
    tis_list = File(
        exists=True, argstr="-TIlist %s", desc="Filename of list of pre-defined TIs."
    )
    t1_list = File(
        exists=True, argstr="-T1list %s", desc="Filename of list of pre-defined T1s"
    )
    t1min = traits.Float(desc="Minimum tissue T1 value [400ms].", argstr="-T1min %f")
    t1max = traits.Float(desc="Maximum tissue T1 value [4000ms].", argstr="-T1max %f")

    # SPGR options
    spgr = traits.Bool(desc="Spoiled Gradient Echo fitting", argstr="-SPGR")
    flips = traits.List(traits.Float, desc="Flip angles", argstr="-flips %s", sep=" ")
    desc = "Filename of list of pre-defined flip angles (deg)."
    flips_list = File(exists=True, argstr="-fliplist %s", desc=desc)
    desc = "Filename of B1 estimate for fitting (or include in prior)."
    b1map = File(exists=True, argstr="-b1map %s", desc=desc)

    # MCMC options:
    mcout = File(
        exists=True, desc="Filename of mc samples (ascii text file)", argstr="-mcout %s"
    )
    mcsamples = traits.Int(
        desc="Number of samples to keep [100].", argstr="-mcsamples %d"
    )
    mcmaxit = traits.Int(
        desc="Number of iterations to run [10,000].", argstr="-mcmaxit %d"
    )
    acceptance = traits.Float(
        desc="Fraction of iterations to accept [0.23].", argstr="-acceptance %f"
    )


class FitQt1OutputSpec(TraitedSpec):
    """Output Spec for FitQt1."""

    t1map_file = File(desc="Filename of the estimated output T1 map (in ms)")
    m0map_file = File(desc="Filename of the m0 map")
    desc = "Filename of the estimated output multi-parameter map"
    mcmap_file = File(desc=desc)
    comp_file = File(desc="Filename of the estimated multi-component T1 map.")
    desc = "Filename of the error map (symmetric matrix, [Diag,OffDiag])"
    error_file = File(desc=desc)
    syn_file = File(desc="Filename of the synthetic ASL data")
    res_file = File(desc="Filename of the model fit residuals")


class FitQt1(NiftyFitCommand):
    """Interface for executable fit_qt1 from Niftyfit platform.

    Use NiftyFit to perform Qt1 fitting.

    T1 Fitting Routine (To inversion recovery or spgr data).
    Fits single component T1 maps in the first instance.

    `Source code <https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyFit-Release>`_

    Examples
    --------

    >>> from nipype.interfaces.niftyfit import FitQt1
    >>> fit_qt1 = FitQt1()
    >>> fit_qt1.inputs.source_file = 'TI4D.nii.gz'
    >>> fit_qt1.cmdline
    'fit_qt1 -source TI4D.nii.gz -comp TI4D_comp.nii.gz \
-error TI4D_error.nii.gz -m0map TI4D_m0map.nii.gz -mcmap TI4D_mcmap.nii.gz \
-res TI4D_res.nii.gz -syn TI4D_syn.nii.gz -t1map TI4D_t1map.nii.gz'

    """

    _cmd = get_custom_path("fit_qt1", env_dir="NIFTYFITDIR")
    input_spec = FitQt1InputSpec
    output_spec = FitQt1OutputSpec
    _suffix = "_fit_qt1"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..registration import MRICoreg


def test_MRICoreg_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        brute_force_limit=dict(
            argstr="--bf-lim %g",
            xor=["no_brute_force"],
        ),
        brute_force_samples=dict(
            argstr="--bf-nsamp %d",
            xor=["no_brute_force"],
        ),
        conform_reference=dict(
            argstr="--conf-ref",
        ),
        dof=dict(
            argstr="--dof %d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ftol=dict(
            argstr="--ftol %e",
        ),
        initial_rotation=dict(
            argstr="--rot %g %g %g",
        ),
        initial_scale=dict(
            argstr="--scale %g %g %g",
        ),
        initial_shear=dict(
            argstr="--shear %g %g %g",
        ),
        initial_translation=dict(
            argstr="--trans %g %g %g",
        ),
        linmintol=dict(
            argstr="--linmintol %e",
        ),
        max_iters=dict(
            argstr="--nitersmax %d",
        ),
        no_brute_force=dict(
            argstr="--no-bf",
        ),
        no_coord_dithering=dict(
            argstr="--no-coord-dither",
        ),
        no_cras0=dict(
            argstr="--no-cras0",
        ),
        no_intensity_dithering=dict(
            argstr="--no-intensity-dither",
        ),
        no_smooth=dict(
            argstr="--no-smooth",
        ),
        num_threads=dict(
            argstr="--threads %d",
        ),
        out_lta_file=dict(
            argstr="--lta %s",
            usedefault=True,
        ),
        out_params_file=dict(
            argstr="--params %s",
        ),
        out_reg_file=dict(
            argstr="--regdat %s",
        ),
        ref_fwhm=dict(
            argstr="--ref-fwhm",
        ),
        reference_file=dict(
            argstr="--ref %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            xor=["subject_id"],
        ),
        reference_mask=dict(
            argstr="--ref-mask %s",
            position=2,
        ),
        saturation_threshold=dict(
            argstr="--sat %g",
        ),
        sep=dict(
            argstr="--sep %s...",
        ),
        source_file=dict(
            argstr="--mov %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        source_mask=dict(
            argstr="--mov-mask",
        ),
        source_oob=dict(
            argstr="--mov-oob",
        ),
        subject_id=dict(
            argstr="--s %s",
            mandatory=True,
            position=1,
            requires=["subjects_dir"],
            xor=["reference_file"],
        ),
        subjects_dir=dict(
            argstr="--sd %s",
        ),
    )
    inputs = MRICoreg.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MRICoreg_outputs():
    output_map = dict(
        out_lta_file=dict(
            extensions=None,
        ),
        out_params_file=dict(
            extensions=None,
        ),
        out_reg_file=dict(
            extensions=None,
        ),
    )
    outputs = MRICoreg.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import Contrast


def test_Contrast_inputs():
    input_map = dict(
        annotation=dict(
            extensions=None,
            mandatory=True,
        ),
        args=dict(
            argstr="%s",
        ),
        copy_inputs=dict(),
        cortex=dict(
            extensions=None,
            mandatory=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        hemisphere=dict(
            argstr="--%s-only",
            mandatory=True,
        ),
        orig=dict(
            extensions=None,
            mandatory=True,
        ),
        rawavg=dict(
            extensions=None,
            mandatory=True,
        ),
        subject_id=dict(
            argstr="--s %s",
            mandatory=True,
            usedefault=True,
        ),
        subjects_dir=dict(),
        thickness=dict(
            extensions=None,
            mandatory=True,
        ),
        white=dict(
            extensions=None,
            mandatory=True,
        ),
    )
    inputs = Contrast.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Contrast_outputs():
    output_map = dict(
        out_contrast=dict(
            extensions=None,
        ),
        out_log=dict(
            extensions=None,
        ),
        out_stats=dict(
            extensions=None,
        ),
    )
    outputs = Contrast.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..segmentation import SimilarityIndex


def test_SimilarityIndex_inputs():
    input_map = dict(
        ANNContinuousVolume=dict(
            argstr="--ANNContinuousVolume %s",
            extensions=None,
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputManualVolume=dict(
            argstr="--inputManualVolume %s",
            extensions=None,
        ),
        outputCSVFilename=dict(
            argstr="--outputCSVFilename %s",
            extensions=None,
        ),
        thresholdInterval=dict(
            argstr="--thresholdInterval %f",
        ),
    )
    inputs = SimilarityIndex.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SimilarityIndex_outputs():
    output_map = dict()
    outputs = SimilarityIndex.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..specialized import EMSegmentCommandLine


def test_EMSegmentCommandLine_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        atlasVolumeFileNames=dict(
            argstr="--atlasVolumeFileNames %s...",
        ),
        disableCompression=dict(
            argstr="--disableCompression ",
        ),
        disableMultithreading=dict(
            argstr="--disableMultithreading %d",
        ),
        dontUpdateIntermediateData=dict(
            argstr="--dontUpdateIntermediateData %d",
        ),
        dontWriteResults=dict(
            argstr="--dontWriteResults ",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        generateEmptyMRMLSceneAndQuit=dict(
            argstr="--generateEmptyMRMLSceneAndQuit %s",
            hash_files=False,
        ),
        intermediateResultsDirectory=dict(
            argstr="--intermediateResultsDirectory %s",
        ),
        keepTempFiles=dict(
            argstr="--keepTempFiles ",
        ),
        loadAtlasNonCentered=dict(
            argstr="--loadAtlasNonCentered ",
        ),
        loadTargetCentered=dict(
            argstr="--loadTargetCentered ",
        ),
        mrmlSceneFileName=dict(
            argstr="--mrmlSceneFileName %s",
            extensions=None,
        ),
        parametersMRMLNodeName=dict(
            argstr="--parametersMRMLNodeName %s",
        ),
        registrationAffineType=dict(
            argstr="--registrationAffineType %d",
        ),
        registrationDeformableType=dict(
            argstr="--registrationDeformableType %d",
        ),
        registrationPackage=dict(
            argstr="--registrationPackage %s",
        ),
        resultMRMLSceneFileName=dict(
            argstr="--resultMRMLSceneFileName %s",
            hash_files=False,
        ),
        resultStandardVolumeFileName=dict(
            argstr="--resultStandardVolumeFileName %s",
            extensions=None,
        ),
        resultVolumeFileName=dict(
            argstr="--resultVolumeFileName %s",
            hash_files=False,
        ),
        targetVolumeFileNames=dict(
            argstr="--targetVolumeFileNames %s...",
        ),
        taskPreProcessingSetting=dict(
            argstr="--taskPreProcessingSetting %s",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
    )
    inputs = EMSegmentCommandLine.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_EMSegmentCommandLine_outputs():
    output_map = dict(
        generateEmptyMRMLSceneAndQuit=dict(
            extensions=None,
        ),
        resultMRMLSceneFileName=dict(
            extensions=None,
        ),
        resultVolumeFileName=dict(
            extensions=None,
        ),
    )
    outputs = EMSegmentCommandLine.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""
3D Slicer is a platform for medical image informatics processing and visualization.

For an EXPERIMENTAL implementation of an interface for the ``3dSlicer`` full framework,
please check `"dynamic" Slicer <nipype.interfaces.dynamic_slicer.html>`__.
"""
from .diffusion import *
from .segmentation import *
from .filtering import *
from .utilities import EMSegmentTransformToNewFormat
from .surface import (
    MergeModels,
    ModelToLabelMap,
    GrayscaleModelMaker,
    ProbeVolumeWithModel,
    LabelMapSmoothing,
    ModelMaker,
)
from .quantification import *
from .legacy import *
from .registration import *
from .converters import DicomToNrrdConverter, OrientScalarVolume



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Nitime is a library for time-series analysis of data from neuroscience experiments."""
from .analysis import (
    CoherenceAnalyzerInputSpec,
    CoherenceAnalyzerOutputSpec,
    CoherenceAnalyzer,
)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..featuredetection import CannySegmentationLevelSetImageFilter


def test_CannySegmentationLevelSetImageFilter_inputs():
    input_map = dict(
        advectionWeight=dict(
            argstr="--advectionWeight %f",
        ),
        args=dict(
            argstr="%s",
        ),
        cannyThreshold=dict(
            argstr="--cannyThreshold %f",
        ),
        cannyVariance=dict(
            argstr="--cannyVariance %f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        initialModel=dict(
            argstr="--initialModel %s",
            extensions=None,
        ),
        initialModelIsovalue=dict(
            argstr="--initialModelIsovalue %f",
        ),
        inputVolume=dict(
            argstr="--inputVolume %s",
            extensions=None,
        ),
        maxIterations=dict(
            argstr="--maxIterations %d",
        ),
        outputSpeedVolume=dict(
            argstr="--outputSpeedVolume %s",
            hash_files=False,
        ),
        outputVolume=dict(
            argstr="--outputVolume %s",
            hash_files=False,
        ),
    )
    inputs = CannySegmentationLevelSetImageFilter.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_CannySegmentationLevelSetImageFilter_outputs():
    output_map = dict(
        outputSpeedVolume=dict(
            extensions=None,
        ),
        outputVolume=dict(
            extensions=None,
        ),
    )
    outputs = CannySegmentationLevelSetImageFilter.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Base interface for nipy """

from ..base import LibraryBaseInterface
from ...utils.misc import package_check

# Originally set in model, preprocess and utils
# Set here to be imported, in case anybody depends on its presence
# Remove in 2.0
have_nipy = True
try:
    package_check("nipy")
except ImportError:
    have_nipy = False


class NipyBaseInterface(LibraryBaseInterface):
    _pkg = "nipy"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..developer import MedicAlgorithmMipavReorient


def test_MedicAlgorithmMipavReorient_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inInterpolation=dict(
            argstr="--inInterpolation %s",
        ),
        inNew=dict(
            argstr="--inNew %s",
        ),
        inResolution=dict(
            argstr="--inResolution %s",
        ),
        inSource=dict(
            argstr="--inSource %s",
            sep=";",
        ),
        inTemplate=dict(
            argstr="--inTemplate %s",
            extensions=None,
        ),
        inUser=dict(
            argstr="--inUser %s",
        ),
        inUser2=dict(
            argstr="--inUser2 %s",
        ),
        inUser3=dict(
            argstr="--inUser3 %s",
        ),
        inUser4=dict(
            argstr="--inUser4 %s",
        ),
        null=dict(
            argstr="--null %s",
        ),
        outReoriented=dict(
            argstr="--outReoriented %s",
            sep=";",
        ),
        xDefaultMem=dict(
            argstr="-xDefaultMem %d",
        ),
        xMaxProcess=dict(
            argstr="-xMaxProcess %d",
            usedefault=True,
        ),
        xPrefExt=dict(
            argstr="--xPrefExt %s",
        ),
    )
    inputs = MedicAlgorithmMipavReorient.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MedicAlgorithmMipavReorient_outputs():
    output_map = dict()
    outputs = MedicAlgorithmMipavReorient.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""This script generates Slicer Interfaces based on the CLI modules XML. CLI
modules are selected from the hardcoded list below and generated code is placed
in the cli_modules.py file (and imported in __init__.py). For this to work
correctly you must have your CLI executabes in $PATH"""
import xml.dom.minidom
import subprocess
import os
from shutil import rmtree

import keyword

python_keywords = (
    keyword.kwlist
)  # If c++ SEM module uses one of these key words as a command line parameter, we need to modify variable


def force_to_valid_python_variable_name(old_name):
    """Valid c++ names are not always valid in python, so
    provide alternate naming

    >>> force_to_valid_python_variable_name('lambda')
    'opt_lambda'
    >>> force_to_valid_python_variable_name('inputVolume')
    'inputVolume'
    """
    new_name = old_name
    new_name = new_name.lstrip().rstrip()
    if old_name in python_keywords:
        new_name = "opt_" + old_name
    return new_name


def add_class_to_package(class_codes, class_names, module_name, package_dir):
    module_python_filename = os.path.join(package_dir, "%s.py" % module_name)
    f_m = open(module_python_filename, "w")
    f_i = open(os.path.join(package_dir, "__init__.py"), "a+")
    f_m.write(
        """# -*- coding: utf-8 -*-
\"\"\"Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator.\"\"\"\n\n"""
    )
    imports = """\
from ..base import (CommandLine, CommandLineInputSpec, SEMLikeCommandLine, TraitedSpec,
                    File, Directory, traits, isdefined, InputMultiPath, OutputMultiPath)
import os\n\n\n"""
    f_m.write(imports)
    f_m.write("\n\n".join(class_codes))
    f_i.write("from %s import %s\n" % (module_name, ", ".join(class_names)))
    f_m.close()
    f_i.close()


def crawl_code_struct(code_struct, package_dir):
    subpackages = []
    for k, v in code_struct.items():
        if isinstance(v, str) or isinstance(v, (str, bytes)):
            module_name = k.lower()
            class_name = k
            class_code = v
            add_class_to_package([class_code], [class_name], module_name, package_dir)
        else:
            l1 = {}
            l2 = {}
            for key in list(v.keys()):
                if isinstance(v[key], str) or isinstance(v[key], (str, bytes)):
                    l1[key] = v[key]
                else:
                    l2[key] = v[key]
            if l2:
                v = l2
                subpackages.append(k.lower())
                f_i = open(os.path.join(package_dir, "__init__.py"), "a+")
                f_i.write("from %s import *\n" % k.lower())
                f_i.close()
                new_pkg_dir = os.path.join(package_dir, k.lower())
                if os.path.exists(new_pkg_dir):
                    rmtree(new_pkg_dir)
                os.mkdir(new_pkg_dir)
                crawl_code_struct(v, new_pkg_dir)
                if l1:
                    for ik, iv in l1.items():
                        crawl_code_struct({ik: {ik: iv}}, new_pkg_dir)
            elif l1:
                v = l1
                module_name = k.lower()
                add_class_to_package(
                    list(v.values()), list(v.keys()), module_name, package_dir
                )
        if subpackages:
            f = open(os.path.join(package_dir, "setup.py"), "w")
            f.write(
                """# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('{pkg_name}', parent_package, top_path)

    {sub_pks}

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())
""".format(
                    pkg_name=package_dir.split("/")[-1],
                    sub_pks="\n    ".join(
                        [
                            "config.add_data_dir('%s')" % sub_pkg
                            for sub_pkg in subpackages
                        ]
                    ),
                )
            )
            f.close()


def generate_all_classes(
    modules_list=[], launcher=[], redirect_x=False, mipav_hacks=False
):
    """modules_list contains all the SEM compliant tools that should have wrappers created for them.
    launcher containtains the command line prefix wrapper arugments needed to prepare
    a proper environment for each of the modules.
    """
    all_code = {}
    for module in modules_list:
        print("=" * 80)
        print("Generating Definition for module {0}".format(module))
        print("^" * 80)
        package, code, module = generate_class(
            module, launcher, redirect_x=redirect_x, mipav_hacks=mipav_hacks
        )
        cur_package = all_code
        module_name = package.strip().split(" ")[0].split(".")[-1]
        for package in package.strip().split(" ")[0].split(".")[:-1]:
            if package not in cur_package:
                cur_package[package] = {}
            cur_package = cur_package[package]
        if module_name not in cur_package:
            cur_package[module_name] = {}
        cur_package[module_name][module] = code
    if os.path.exists("__init__.py"):
        os.unlink("__init__.py")
    crawl_code_struct(all_code, os.getcwd())


def generate_class(
    module, launcher, strip_module_name_prefix=True, redirect_x=False, mipav_hacks=False
):
    dom = grab_xml(module, launcher, mipav_hacks=mipav_hacks)
    if strip_module_name_prefix:
        module_name = module.split(".")[-1]
    else:
        module_name = module
    inputTraits = []
    outputTraits = []
    outputs_filenames = {}

    # self._outputs_nodes = []

    class_string = '"""'

    for desc_str in [
        "title",
        "category",
        "description",
        "version",
        "documentation-url",
        "license",
        "contributor",
        "acknowledgements",
    ]:
        el = dom.getElementsByTagName(desc_str)
        if el and el[0].firstChild and el[0].firstChild.nodeValue.strip():
            class_string += (
                desc_str + ": " + el[0].firstChild.nodeValue.strip() + "\n\n"
            )
        if desc_str == "category":
            category = el[0].firstChild.nodeValue.strip()
    class_string += '"""'

    for paramGroup in dom.getElementsByTagName("parameters"):
        indices = paramGroup.getElementsByTagName("index")
        max_index = 0
        for index in indices:
            if int(index.firstChild.nodeValue) > max_index:
                max_index = int(index.firstChild.nodeValue)
        for param in paramGroup.childNodes:
            if param.nodeName in ["label", "description", "#text", "#comment"]:
                continue
            traitsParams = {}

            longFlagNode = param.getElementsByTagName("longflag")
            if longFlagNode:
                # Prefer to use longFlag as name if it is given, rather than the parameter name
                longFlagName = longFlagNode[0].firstChild.nodeValue
                # SEM automatically strips prefixed "--" or "-" from from xml before processing
                # we need to replicate that behavior here The following
                # two nodes in xml have the same behavior in the program
                # <longflag>--test</longflag>
                # <longflag>test</longflag>
                longFlagName = longFlagName.lstrip(" -").rstrip(" ")
                name = longFlagName
                name = force_to_valid_python_variable_name(name)
                traitsParams["argstr"] = "--" + longFlagName + " "
            else:
                name = param.getElementsByTagName("name")[0].firstChild.nodeValue
                name = force_to_valid_python_variable_name(name)
                if param.getElementsByTagName("index"):
                    traitsParams["argstr"] = ""
                else:
                    traitsParams["argstr"] = "--" + name + " "

            if (
                param.getElementsByTagName("description")
                and param.getElementsByTagName("description")[0].firstChild
            ):
                traitsParams["desc"] = (
                    param.getElementsByTagName("description")[0]
                    .firstChild.nodeValue.replace('"', '\\"')
                    .replace("\n", ", ")
                )

            argsDict = {
                "directory": "%s",
                "file": "%s",
                "integer": "%d",
                "double": "%f",
                "float": "%f",
                "image": "%s",
                "transform": "%s",
                "boolean": "",
                "string-enumeration": "%s",
                "string": "%s",
                "integer-enumeration": "%s",
                "table": "%s",
                "point": "%s",
                "region": "%s",
                "geometry": "%s",
            }

            if param.nodeName.endswith("-vector"):
                traitsParams["argstr"] += "%s"
            else:
                traitsParams["argstr"] += argsDict[param.nodeName]

            index = param.getElementsByTagName("index")
            if index:
                traitsParams["position"] = int(index[0].firstChild.nodeValue) - (
                    max_index + 1
                )

            desc = param.getElementsByTagName("description")
            if index:
                traitsParams["desc"] = desc[0].firstChild.nodeValue

            typesDict = {
                "integer": "traits.Int",
                "double": "traits.Float",
                "float": "traits.Float",
                "image": "File",
                "transform": "File",
                "boolean": "traits.Bool",
                "string": "traits.Str",
                "file": "File",
                "geometry": "File",
                "directory": "Directory",
                "table": "File",
                "point": "traits.List",
                "region": "traits.List",
            }

            if param.nodeName.endswith("-enumeration"):
                type = "traits.Enum"
                values = [
                    '"%s"' % str(el.firstChild.nodeValue).replace('"', "")
                    for el in param.getElementsByTagName("element")
                ]
            elif param.nodeName.endswith("-vector"):
                type = "InputMultiPath"
                if param.nodeName in [
                    "file",
                    "directory",
                    "image",
                    "geometry",
                    "transform",
                    "table",
                ]:
                    values = [
                        "%s(exists=True)"
                        % typesDict[param.nodeName.replace("-vector", "")]
                    ]
                else:
                    values = [typesDict[param.nodeName.replace("-vector", "")]]
                if mipav_hacks is True:
                    traitsParams["sep"] = ";"
                else:
                    traitsParams["sep"] = ","
            elif param.getAttribute("multiple") == "true":
                type = "InputMultiPath"
                if param.nodeName in [
                    "file",
                    "directory",
                    "image",
                    "geometry",
                    "transform",
                    "table",
                ]:
                    values = ["%s(exists=True)" % typesDict[param.nodeName]]
                elif param.nodeName in ["point", "region"]:
                    values = [
                        "%s(traits.Float(), minlen=3, maxlen=3)"
                        % typesDict[param.nodeName]
                    ]
                else:
                    values = [typesDict[param.nodeName]]
                traitsParams["argstr"] += "..."
            else:
                values = []
                type = typesDict[param.nodeName]

            if param.nodeName in [
                "file",
                "directory",
                "image",
                "geometry",
                "transform",
                "table",
            ]:
                if not param.getElementsByTagName("channel"):
                    raise RuntimeError(
                        "Insufficient XML specification: each element of type 'file', 'directory', 'image', 'geometry', 'transform',  or 'table' requires 'channel' field.\n{0}".format(
                            traitsParams
                        )
                    )
                elif (
                    param.getElementsByTagName("channel")[0].firstChild.nodeValue
                    == "output"
                ):
                    traitsParams["hash_files"] = False
                    inputTraits.append(
                        "%s = traits.Either(traits.Bool, %s(%s), %s)"
                        % (
                            name,
                            type,
                            parse_values(values).replace("exists=True", ""),
                            parse_params(traitsParams),
                        )
                    )
                    traitsParams["exists"] = True
                    traitsParams.pop("argstr")
                    traitsParams.pop("hash_files")
                    outputTraits.append(
                        "%s = %s(%s%s)"
                        % (
                            name,
                            type.replace("Input", "Output"),
                            parse_values(values),
                            parse_params(traitsParams),
                        )
                    )

                    outputs_filenames[name] = gen_filename_from_param(param, name)
                elif (
                    param.getElementsByTagName("channel")[0].firstChild.nodeValue
                    == "input"
                ):
                    if (
                        param.nodeName
                        in [
                            "file",
                            "directory",
                            "image",
                            "geometry",
                            "transform",
                            "table",
                        ]
                        and type not in ["InputMultiPath", "traits.List"]
                    ):
                        traitsParams["exists"] = True
                    inputTraits.append(
                        "%s = %s(%s%s)"
                        % (name, type, parse_values(values), parse_params(traitsParams))
                    )
                else:
                    raise RuntimeError(
                        "Insufficient XML specification: each element of type 'file', 'directory', 'image', 'geometry', 'transform',  or 'table' requires 'channel' field to be in ['input','output'].\n{0}".format(
                            traitsParams
                        )
                    )
            else:  # For all other parameter types, they are implicitly only input types
                inputTraits.append(
                    "%s = %s(%s%s)"
                    % (name, type, parse_values(values), parse_params(traitsParams))
                )

    if mipav_hacks:
        blacklisted_inputs = ["maxMemoryUsage"]
        inputTraits = [
            trait for trait in inputTraits if trait.split()[0] not in blacklisted_inputs
        ]

        compulsory_inputs = [
            'xDefaultMem = traits.Int(desc="Set default maximum heap size", argstr="-xDefaultMem %d")',
            'xMaxProcess = traits.Int(1, desc="Set default maximum number of processes.", argstr="-xMaxProcess %d", usedefault=True)',
        ]
        inputTraits += compulsory_inputs

    input_spec_code = "class " + module_name + "InputSpec(CommandLineInputSpec):\n"
    for trait in inputTraits:
        input_spec_code += "    " + trait + "\n"

    output_spec_code = "class " + module_name + "OutputSpec(TraitedSpec):\n"
    if not outputTraits:
        output_spec_code += "    pass\n"
    else:
        for trait in outputTraits:
            output_spec_code += "    " + trait + "\n"

    output_filenames_code = "_outputs_filenames = {"
    output_filenames_code += ",".join(
        ["'%s':'%s'" % (key, value) for key, value in outputs_filenames.items()]
    )
    output_filenames_code += "}"

    input_spec_code += "\n\n"
    output_spec_code += "\n\n"

    template = """class %module_name%(SEMLikeCommandLine):
    %class_str%

    input_spec = %module_name%InputSpec
    output_spec = %module_name%OutputSpec
    _cmd = "%launcher% %name% "
    %output_filenames_code%\n"""
    template += "    _redirect_x = {0}\n".format(str(redirect_x))

    main_class = (
        template.replace("%class_str%", class_string)
        .replace("%module_name%", module_name)
        .replace("%name%", module)
        .replace("%output_filenames_code%", output_filenames_code)
        .replace("%launcher%", " ".join(launcher))
    )

    return category, input_spec_code + output_spec_code + main_class, module_name


def grab_xml(module, launcher, mipav_hacks=False):
    #        cmd = CommandLine(command = "Slicer3", args="--launch %s --xml"%module)
    #        ret = cmd.run()
    command_list = launcher[:]  # force copy to preserve original
    command_list.extend([module, "--xml"])
    final_command = " ".join(command_list)
    xmlReturnValue = subprocess.Popen(
        final_command, stdout=subprocess.PIPE, shell=True
    ).communicate()[0]
    if mipav_hacks:
        # workaround for a jist bug https://www.nitrc.org/tracker/index.php?func=detail&aid=7234&group_id=228&atid=942
        new_xml = ""
        replace_closing_tag = False
        for line in xmlReturnValue.splitlines():
            if line.strip() == "<file collection: semi-colon delimited list>":
                new_xml += "<file-vector>\n"
                replace_closing_tag = True
            elif replace_closing_tag and line.strip() == "</file>":
                new_xml += "</file-vector>\n"
                replace_closing_tag = False
            else:
                new_xml += line + "\n"

        xmlReturnValue = new_xml

        # workaround for a JIST bug https://www.nitrc.org/tracker/index.php?func=detail&aid=7233&group_id=228&atid=942
        if xmlReturnValue.strip().endswith("XML"):
            xmlReturnValue = xmlReturnValue.strip()[:-3]
        if xmlReturnValue.strip().startswith("Error: Unable to set default atlas"):
            xmlReturnValue = xmlReturnValue.strip()[
                len("Error: Unable to set default atlas") :
            ]
    try:
        dom = xml.dom.minidom.parseString(xmlReturnValue.strip())
    except Exception as e:
        print(xmlReturnValue.strip())
        raise e
    return dom


#        if ret.runtime.returncode == 0:
#            return xml.dom.minidom.parseString(ret.runtime.stdout)
#        else:
#            raise Exception(cmd.cmdline + " failed:\n%s"%ret.runtime.stderr)


def parse_params(params):
    list = []
    for key, value in params.items():
        if isinstance(value, (str, bytes)):
            list.append('%s="%s"' % (key, value.replace('"', "'")))
        else:
            list.append("%s=%s" % (key, value))

    return ", ".join(list)


def parse_values(values):
    values = ["%s" % value for value in values]
    if len(values) > 0:
        retstr = ", ".join(values) + ", "
    else:
        retstr = ""
    return retstr


def gen_filename_from_param(param, base):
    fileExtensions = param.getAttribute("fileExtensions")
    if fileExtensions:
        # It is possible that multiple file extensions can be specified in a
        # comma separated list,  This will extract just the first extension
        firstFileExtension = fileExtensions.split(",")[0]
        ext = firstFileExtension
    else:
        ext = {
            "image": ".nii",
            "transform": ".mat",
            "file": "",
            "directory": "",
            "geometry": ".vtk",
        }[param.nodeName]
    return base + ext


if __name__ == "__main__":
    # NOTE:  For now either the launcher needs to be found on the default path, or
    # every tool in the modules list must be found on the default path
    # AND calling the module with --xml must be supported and compliant.
    modules_list = [
        "MedianImageFilter",
        "CheckerBoardFilter",
        "EMSegmentCommandLine",
        "GrayscaleFillHoleImageFilter",
        # 'CreateDICOMSeries', #missing channel
        "TractographyLabelMapSeeding",
        "IntensityDifferenceMetric",
        "DWIToDTIEstimation",
        "MaskScalarVolume",
        "ImageLabelCombine",
        "DTIimport",
        "OtsuThresholdImageFilter",
        "ExpertAutomatedRegistration",
        "ThresholdScalarVolume",
        "DWIUnbiasedNonLocalMeansFilter",
        "BRAINSFit",
        "MergeModels",
        "ResampleDTIVolume",
        "MultiplyScalarVolumes",
        "LabelMapSmoothing",
        "RigidRegistration",
        "VotingBinaryHoleFillingImageFilter",
        "BRAINSROIAuto",
        "RobustStatisticsSegmenter",
        "GradientAnisotropicDiffusion",
        "ProbeVolumeWithModel",
        "ModelMaker",
        "ExtractSkeleton",
        "GrayscaleGrindPeakImageFilter",
        "N4ITKBiasFieldCorrection",
        "BRAINSResample",
        "DTIexport",
        "VBRAINSDemonWarp",
        "ResampleScalarVectorDWIVolume",
        "ResampleScalarVolume",
        "OtsuThresholdSegmentation",
        # 'ExecutionModelTour',
        "HistogramMatching",
        "BRAINSDemonWarp",
        "ModelToLabelMap",
        "GaussianBlurImageFilter",
        "DiffusionWeightedVolumeMasking",
        "GrayscaleModelMaker",
        "CastScalarVolume",
        "DicomToNrrdConverter",
        "AffineRegistration",
        "AddScalarVolumes",
        "LinearRegistration",
        "SimpleRegionGrowingSegmentation",
        "DWIJointRicianLMMSEFilter",
        "MultiResolutionAffineRegistration",
        "SubtractScalarVolumes",
        "DWIRicianLMMSEFilter",
        "OrientScalarVolume",
        "FiducialRegistration",
        "BSplineDeformableRegistration",
        "CurvatureAnisotropicDiffusion",
        "PETStandardUptakeValueComputation",
        "DiffusionTensorScalarMeasurements",
        "ACPCTransform",
        "EMSegmentTransformToNewFormat",
        "BSplineToDeformationField",
    ]

    # SlicerExecutionModel compliant tools that are usually statically built, and don't need the Slicer3 --launcher
    generate_all_classes(modules_list=modules_list, launcher=[])
    # Tools compliant with SlicerExecutionModel called from the Slicer environment (for shared lib compatibility)
    # launcher = ['/home/raid3/gorgolewski/software/slicer/Slicer', '--launch']
    # generate_all_classes(modules_list=modules_list, launcher=launcher)
    # generate_all_classes(modules_list=['BRAINSABC'], launcher=[] )



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..bru2nii import Bru2


def test_Bru2_inputs():
    input_map = dict(
        actual_size=dict(
            argstr="-a",
        ),
        append_protocol_name=dict(
            argstr="-p",
        ),
        args=dict(
            argstr="%s",
        ),
        compress=dict(
            argstr="-z",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        force_conversion=dict(
            argstr="-f",
        ),
        input_dir=dict(
            argstr="%s",
            mandatory=True,
            position=-1,
        ),
        output_filename=dict(
            argstr="-o %s",
            genfile=True,
        ),
    )
    inputs = Bru2.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Bru2_outputs():
    output_map = dict(
        nii_file=dict(
            extensions=None,
        ),
    )
    outputs = Bru2.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..regutils import RegTools


def test_RegTools_inputs():
    input_map = dict(
        add_val=dict(
            argstr="-add %s",
        ),
        args=dict(
            argstr="%s",
        ),
        bin_flag=dict(
            argstr="-bin",
        ),
        chg_res_val=dict(
            argstr="-chgres %f %f %f",
        ),
        div_val=dict(
            argstr="-div %s",
        ),
        down_flag=dict(
            argstr="-down",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-in %s",
            extensions=None,
            mandatory=True,
        ),
        inter_val=dict(
            argstr="-interp %d",
        ),
        iso_flag=dict(
            argstr="-iso",
        ),
        mask_file=dict(
            argstr="-nan %s",
            extensions=None,
        ),
        mul_val=dict(
            argstr="-mul %s",
        ),
        noscl_flag=dict(
            argstr="-noscl",
        ),
        omp_core_val=dict(
            argstr="-omp %i",
            usedefault=True,
        ),
        out_file=dict(
            argstr="-out %s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_tools.nii.gz",
        ),
        rms_val=dict(
            argstr="-rms %s",
            extensions=None,
        ),
        smo_g_val=dict(
            argstr="-smoG %f %f %f",
        ),
        smo_s_val=dict(
            argstr="-smoS %f %f %f",
        ),
        sub_val=dict(
            argstr="-sub %s",
        ),
        thr_val=dict(
            argstr="-thr %f",
        ),
    )
    inputs = RegTools.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_RegTools_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = RegTools.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# Use duecredit (duecredit.org) to provide a citation to relevant work to
# be cited. This does nothing, unless the user has duecredit installed,
# And calls this with duecredit (as in `python -m duecredit script.py`):
from .external.due import due, Doi, BibTeX

due.cite(
    Doi("10.3389/fninf.2011.00013"),
    description="A flexible, lightweight and extensible neuroimaging data"
    " processing framework in Python",
    path="nipype",
    tags=["implementation"],
)

due.cite(
    Doi("10.5281/zenodo.50186"),
    description="A flexible, lightweight and extensible neuroimaging data"
    " processing framework in Python",
    path="nipype",
    tags=["implementation"],
)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Created on 20 Apr 2010

logging options : INFO, DEBUG
hash_method : content, timestamp

@author: Chris Filo Gorgolewski
"""
import os
import sys
import errno
import atexit
from warnings import warn
from distutils.version import LooseVersion
import configparser
import numpy as np

from simplejson import load, dump

from .misc import str2bool
from filelock import SoftFileLock

CONFIG_DEPRECATIONS = {
    "profile_runtime": ("monitoring.enabled", "1.0"),
    "filemanip_level": ("logging.utils_level", "1.0"),
}


DEFAULT_CONFIG_TPL = """\
[logging]
workflow_level = INFO
utils_level = INFO
interface_level = INFO
log_to_file = false
log_directory = {log_dir}
log_size = 16384000
log_rotate = 4

[execution]
create_report = true
crashdump_dir = {crashdump_dir}
hash_method = timestamp
job_finished_timeout = 5
keep_inputs = false
local_hash_check = true
matplotlib_backend = Agg
plugin = Linear
remove_node_directories = false
remove_unnecessary_outputs = true
try_hard_link_datasink = true
single_thread_matlab = true
crashfile_format = pklz
stop_on_first_crash = false
stop_on_first_rerun = false
use_relative_paths = false
stop_on_unknown_version = false
write_provenance = false
parameterize_dirs = true
poll_sleep_duration = 2
xvfb_max_wait = 10
check_version = true

[monitoring]
enabled = false
sample_frequency = 1
summary_append = true

[check]
interval = 1209600
"""


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class NipypeConfig(object):
    """Base nipype config class"""

    def __init__(self, *args, **kwargs):
        self._config = configparser.ConfigParser()
        self._cwd = None

        config_dir = os.path.expanduser(
            os.getenv("NIPYPE_CONFIG_DIR", default="~/.nipype")
        )
        self.data_file = os.path.join(config_dir, "nipype.json")

        self.set_default_config()
        self._display = None
        self._resource_monitor = None

        self._config.read([os.path.join(config_dir, "nipype.cfg"), "nipype.cfg"])

        for option in CONFIG_DEPRECATIONS:
            for section in ["execution", "logging", "monitoring"]:
                if self.has_option(section, option):
                    new_section, new_option = CONFIG_DEPRECATIONS[option][0].split(".")
                    if not self.has_option(new_section, new_option):
                        # Warn implicit in get
                        self.set(new_section, new_option, self.get(section, option))

    @property
    def cwd(self):
        """Cache current working directory ASAP"""
        # Run getcwd only once, preventing multiproc to finish
        # with error having changed to the wrong path
        if self._cwd is None:
            try:
                self._cwd = os.getcwd()
            except OSError:
                warn(
                    'Trying to run Nipype from a nonexistent directory "{}".'.format(
                        os.getenv("PWD", "unknown")
                    ),
                    RuntimeWarning,
                )
                raise
        return self._cwd

    def set_default_config(self):
        """Read default settings template and set into config object"""
        default_cfg = DEFAULT_CONFIG_TPL.format(
            log_dir=os.path.expanduser("~"),  # Get $HOME in a platform-agnostic way
            crashdump_dir=self.cwd,  # Read cached cwd
        )

        try:
            self._config.read_string(default_cfg)  # Python >= 3.2
        except AttributeError:
            from io import StringIO

            self._config.readfp(StringIO(default_cfg))

    def enable_debug_mode(self):
        """Enables debug configuration"""
        from .. import logging

        self._config.set("execution", "stop_on_first_crash", "true")
        self._config.set("execution", "remove_unnecessary_outputs", "false")
        self._config.set("execution", "keep_inputs", "true")
        self._config.set("logging", "workflow_level", "DEBUG")
        self._config.set("logging", "interface_level", "DEBUG")
        self._config.set("logging", "utils_level", "DEBUG")
        logging.update_logging(self._config)

    def set_log_dir(self, log_dir):
        """Sets logging directory

        This should be the first thing that is done before any nipype class
        with logging is imported.
        """
        self._config.set("logging", "log_directory", log_dir)

    def get(self, section, option, default=None):
        """Get an option"""
        if option in CONFIG_DEPRECATIONS:
            msg = (
                'Config option "%s" has been deprecated as of nipype %s. '
                'Please use "%s" instead.'
            ) % (option, CONFIG_DEPRECATIONS[option][1], CONFIG_DEPRECATIONS[option][0])
            warn(msg)
            section, option = CONFIG_DEPRECATIONS[option][0].split(".")

        if self._config.has_option(section, option):
            return self._config.get(section, option)
        return default

    def set(self, section, option, value):
        """Set new value on option"""
        if isinstance(value, bool):
            value = str(value)

        if option in CONFIG_DEPRECATIONS:
            msg = (
                'Config option "%s" has been deprecated as of nipype %s. '
                'Please use "%s" instead.'
            ) % (option, CONFIG_DEPRECATIONS[option][1], CONFIG_DEPRECATIONS[option][0])
            warn(msg)
            section, option = CONFIG_DEPRECATIONS[option][0].split(".")

        return self._config.set(section, option, value)

    def getboolean(self, section, option):
        """Get a boolean option from section"""
        return self._config.getboolean(section, option)

    def has_option(self, section, option):
        """Check if option exists in section"""
        return self._config.has_option(section, option)

    @property
    def _sections(self):
        return self._config._sections

    def get_data(self, key):
        """Read options file"""
        if not os.path.exists(self.data_file):
            return None
        with SoftFileLock("%s.lock" % self.data_file):
            with open(self.data_file, "rt") as file:
                datadict = load(file)
        if key in datadict:
            return datadict[key]
        return None

    def save_data(self, key, value):
        """Store config flie"""
        datadict = {}
        if os.path.exists(self.data_file):
            with SoftFileLock("%s.lock" % self.data_file):
                with open(self.data_file, "rt") as file:
                    datadict = load(file)
        else:
            dirname = os.path.dirname(self.data_file)
            if not os.path.exists(dirname):
                mkdir_p(dirname)
        with SoftFileLock("%s.lock" % self.data_file):
            with open(self.data_file, "wt") as file:
                datadict[key] = value
                dump(datadict, file)

    def update_config(self, config_dict):
        """Extend internal dictionary with config_dict"""
        for section in ["execution", "logging", "check"]:
            if section in config_dict:
                for key, val in list(config_dict[section].items()):
                    if not key.startswith("__"):
                        self._config.set(section, key, str(val))

    def update_matplotlib(self):
        """Set backend on matplotlib from options"""
        import matplotlib

        matplotlib.use(self.get("execution", "matplotlib_backend"))

    def enable_provenance(self):
        """Sets provenance storing on"""
        self._config.set("execution", "write_provenance", "true")
        self._config.set("execution", "hash_method", "content")

    @property
    def resource_monitor(self):
        """Check if resource_monitor is available"""
        if self._resource_monitor is not None:
            return self._resource_monitor

        # Cache config from nipype config
        self.resource_monitor = (
            str2bool(self._config.get("monitoring", "enabled")) or False
        )
        return self._resource_monitor

    @resource_monitor.setter
    def resource_monitor(self, value):
        # Accept string true/false values
        if isinstance(value, (str, bytes)):
            value = str2bool(value.lower())

        if value is False:
            self._resource_monitor = False
        elif value is True:
            if not self._resource_monitor:
                # Before setting self._resource_monitor check psutil
                # availability
                self._resource_monitor = False
                try:
                    import psutil

                    self._resource_monitor = LooseVersion(
                        psutil.__version__
                    ) >= LooseVersion("5.0")
                except ImportError:
                    pass
                finally:
                    if not self._resource_monitor:
                        warn(
                            "Could not enable the resource monitor: "
                            "psutil>=5.0 could not be imported."
                        )
                    self._config.set(
                        "monitoring", "enabled", ("%s" % self._resource_monitor).lower()
                    )

    def enable_resource_monitor(self):
        """Sets the resource monitor on"""
        self.resource_monitor = True

    def disable_resource_monitor(self):
        """Sets the resource monitor off"""
        self.resource_monitor = False

    def get_display(self):
        """Returns the first display available"""

        # Check if an Xorg server is listening
        # import subprocess as sp
        # if not hasattr(sp, 'DEVNULL'):
        #     setattr(sp, 'DEVNULL', os.devnull)
        # x_listening = bool(sp.call('ps au | grep -v grep | grep -i xorg',
        #                    shell=True, stdout=sp.DEVNULL))

        if self._display is not None:
            return ":%d" % self._display.new_display

        sysdisplay = None
        if self._config.has_option("execution", "display_variable"):
            sysdisplay = self._config.get("execution", "display_variable")

        sysdisplay = sysdisplay or os.getenv("DISPLAY")
        if sysdisplay:
            from collections import namedtuple

            def _mock():
                pass

            # Store a fake Xvfb object. Format - <host>:<display>[.<screen>]
            ndisp = sysdisplay.split(":")[-1].split(".")[0]
            Xvfb = namedtuple("Xvfb", ["new_display", "stop"])
            self._display = Xvfb(int(ndisp), _mock)
            return self.get_display()
        else:
            if "darwin" in sys.platform:
                raise RuntimeError(
                    "Xvfb requires root permissions to run in OSX. Please "
                    "make sure that an X server is listening and set the "
                    "appropriate config on either $DISPLAY or nipype's "
                    '"display_variable" config. Valid X servers include '
                    "VNC, XQuartz, or manually started Xvfb."
                )

            # If $DISPLAY is empty, it confuses Xvfb so unset
            if sysdisplay == "":
                del os.environ["DISPLAY"]
            try:
                from xvfbwrapper import Xvfb
            except ImportError:
                raise RuntimeError(
                    "A display server was required, but $DISPLAY is not "
                    "defined and Xvfb could not be imported."
                )

            self._display = Xvfb(nolisten="tcp")
            self._display.start()

            # Older versions of xvfbwrapper used vdisplay_num
            if not hasattr(self._display, "new_display"):
                setattr(self._display, "new_display", self._display.vdisplay_num)
            return self.get_display()

    def stop_display(self):
        """Closes the display if started"""
        if self._display is not None:
            from .. import logging

            self._display.stop()
            logging.getLogger("nipype.interface").debug("Closing display (if virtual)")


@atexit.register
def free_display():
    """Stop virtual display (if it is up)"""
    from .. import config

    config.stop_display()



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..thresholdscalarvolume import ThresholdScalarVolume


def test_ThresholdScalarVolume_inputs():
    input_map = dict(
        InputVolume=dict(
            argstr="%s",
            extensions=None,
            position=-2,
        ),
        OutputVolume=dict(
            argstr="%s",
            hash_files=False,
            position=-1,
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        lower=dict(
            argstr="--lower %d",
        ),
        outsidevalue=dict(
            argstr="--outsidevalue %d",
        ),
        threshold=dict(
            argstr="--threshold %d",
        ),
        thresholdtype=dict(
            argstr="--thresholdtype %s",
        ),
        upper=dict(
            argstr="--upper %d",
        ),
    )
    inputs = ThresholdScalarVolume.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ThresholdScalarVolume_outputs():
    output_map = dict(
        OutputVolume=dict(
            extensions=None,
            position=-1,
        ),
    )
    outputs = ThresholdScalarVolume.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import SpaceTimeRealigner


def test_SpaceTimeRealigner_inputs():
    input_map = dict(
        in_file=dict(
            mandatory=True,
            min_ver="0.4.0.dev",
        ),
        slice_info=dict(
            requires=["slice_times"],
        ),
        slice_times=dict(),
        tr=dict(
            requires=["slice_times"],
        ),
    )
    inputs = SpaceTimeRealigner.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SpaceTimeRealigner_outputs():
    output_map = dict(
        out_file=dict(),
        par_file=dict(),
    )
    outputs = SpaceTimeRealigner.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""AFNI's svm interfaces."""

from ..base import TraitedSpec, traits, File
from .base import AFNICommand, AFNICommandInputSpec, AFNICommandOutputSpec


class SVMTrainInputSpec(AFNICommandInputSpec):
    # training options
    ttype = traits.Str(
        desc="tname: classification or regression", argstr="-type %s", mandatory=True
    )
    in_file = File(
        desc="A 3D+t AFNI brik dataset to be used for training.",
        argstr="-trainvol %s",
        mandatory=True,
        exists=True,
        copyfile=False,
    )
    out_file = File(
        name_template="%s_vectors",
        desc="output sum of weighted linear support vectors file name",
        argstr="-bucket %s",
        suffix="_bucket",
        name_source="in_file",
    )
    model = File(
        name_template="%s_model",
        desc="basename for the brik containing the SVM model",
        argstr="-model %s",
        suffix="_model",
        name_source="in_file",
    )
    alphas = File(
        name_template="%s_alphas",
        desc="output alphas file name",
        argstr="-alpha %s",
        suffix="_alphas",
        name_source="in_file",
    )
    mask = File(
        desc="byte-format brik file used to mask voxels in the analysis",
        argstr="-mask %s",
        position=-1,
        exists=True,
        copyfile=False,
    )
    nomodelmask = traits.Bool(
        desc="Flag to enable the omission of a mask file", argstr="-nomodelmask"
    )
    trainlabels = File(
        desc=".1D labels corresponding to the stimulus paradigm for the training data.",
        argstr="-trainlabels %s",
        exists=True,
    )
    censor = File(
        desc=".1D censor file that allows the user to ignore certain samples in the training data.",
        argstr="-censor %s",
        exists=True,
    )
    kernel = traits.Str(
        desc="string specifying type of kernel function:linear, polynomial, rbf, sigmoid",
        argstr="-kernel %s",
    )
    max_iterations = traits.Int(
        desc="Specify the maximum number of iterations for the optimization.",
        argstr="-max_iterations %d",
    )
    w_out = traits.Bool(
        desc="output sum of weighted linear support vectors", argstr="-wout"
    )
    options = traits.Str(desc="additional options for SVM-light", argstr="%s")


class SVMTrainOutputSpec(TraitedSpec):
    out_file = File(desc="sum of weighted linear support vectors file name")
    model = File(desc="brik containing the SVM model file name")
    alphas = File(desc="output alphas file name")


class SVMTrain(AFNICommand):
    """Temporally predictive modeling with the support vector machine
    SVM Train Only
    For complete details, see the `3dsvm Documentation.
    <https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dsvm.html>`_

    Examples
    ========

    >>> from nipype.interfaces import afni as afni
    >>> svmTrain = afni.SVMTrain()
    >>> svmTrain.inputs.in_file = 'run1+orig'
    >>> svmTrain.inputs.trainlabels = 'run1_categories.1D'
    >>> svmTrain.inputs.ttype = 'regression'
    >>> svmTrain.inputs.mask = 'mask.nii'
    >>> svmTrain.inputs.model = 'model_run1'
    >>> svmTrain.inputs.alphas = 'alphas_run1'
    >>> res = svmTrain.run() # doctest: +SKIP

    """

    _cmd = "3dsvm"
    input_spec = SVMTrainInputSpec
    output_spec = SVMTrainOutputSpec
    _additional_metadata = ["suffix"]

    def _format_arg(self, name, trait_spec, value):
        return super(SVMTrain, self)._format_arg(name, trait_spec, value)


class SVMTestInputSpec(AFNICommandInputSpec):
    # testing options
    model = traits.Str(
        desc="modname is the basename for the brik containing the SVM model",
        argstr="-model %s",
        mandatory=True,
    )
    in_file = File(
        desc="A 3D or 3D+t AFNI brik dataset to be used for testing.",
        argstr="-testvol %s",
        exists=True,
        mandatory=True,
    )
    out_file = File(
        name_template="%s_predictions",
        desc="filename for .1D prediction file(s).",
        argstr="-predictions %s",
    )
    testlabels = File(
        desc="*true* class category .1D labels for the test dataset. It is used to calculate the prediction accuracy performance",
        exists=True,
        argstr="-testlabels %s",
    )
    classout = traits.Bool(
        desc="Flag to specify that pname files should be integer-valued, corresponding to class category decisions.",
        argstr="-classout",
    )
    nopredcensord = traits.Bool(
        desc="Flag to prevent writing predicted values for censored time-points",
        argstr="-nopredcensord",
    )
    nodetrend = traits.Bool(
        desc="Flag to specify that pname files should not be linearly detrended",
        argstr="-nodetrend",
    )
    multiclass = traits.Bool(
        desc="Specifies multiclass algorithm for classification",
        argstr="-multiclass %s",
    )
    options = traits.Str(desc="additional options for SVM-light", argstr="%s")


class SVMTest(AFNICommand):
    """Temporally predictive modeling with the support vector machine
    SVM Test Only
    For complete details, see the `3dsvm Documentation.
    <https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dsvm.html>`_

    Examples
    ========

    >>> from nipype.interfaces import afni as afni
    >>> svmTest = afni.SVMTest()
    >>> svmTest.inputs.in_file= 'run2+orig'
    >>> svmTest.inputs.model= 'run1+orig_model'
    >>> svmTest.inputs.testlabels= 'run2_categories.1D'
    >>> svmTest.inputs.out_file= 'pred2_model1'
    >>> res = svmTest.run() # doctest: +SKIP

    """

    _cmd = "3dsvm"
    input_spec = SVMTestInputSpec
    output_spec = AFNICommandOutputSpec



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import EditTransform


def test_EditTransform_inputs():
    input_map = dict(
        interpolation=dict(
            argstr="FinalBSplineInterpolationOrder",
            usedefault=True,
        ),
        output_file=dict(
            extensions=None,
        ),
        output_format=dict(
            argstr="ResultImageFormat",
        ),
        output_type=dict(
            argstr="ResultImagePixelType",
        ),
        reference_image=dict(
            extensions=None,
        ),
        transform_file=dict(
            extensions=None,
            mandatory=True,
        ),
    )
    inputs = EditTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_EditTransform_outputs():
    output_map = dict(
        output_file=dict(
            extensions=None,
        ),
    )
    outputs = EditTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..dti import VecReg


def test_VecReg_inputs():
    input_map = dict(
        affine_mat=dict(
            argstr="-t %s",
            extensions=None,
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        interpolation=dict(
            argstr="--interp=%s",
        ),
        mask=dict(
            argstr="-m %s",
            extensions=None,
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
        output_type=dict(),
        ref_mask=dict(
            argstr="--refmask=%s",
            extensions=None,
        ),
        ref_vol=dict(
            argstr="-r %s",
            extensions=None,
            mandatory=True,
        ),
        rotation_mat=dict(
            argstr="--rotmat=%s",
            extensions=None,
        ),
        rotation_warp=dict(
            argstr="--rotwarp=%s",
            extensions=None,
        ),
        warp_field=dict(
            argstr="-w %s",
            extensions=None,
        ),
    )
    inputs = VecReg.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_VecReg_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = VecReg.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import Coregister


def test_Coregister_inputs():
    input_map = dict(
        apply_to_files=dict(
            copyfile=True,
            field="other",
        ),
        cost_function=dict(
            field="eoptions.cost_fun",
        ),
        fwhm=dict(
            field="eoptions.fwhm",
        ),
        jobtype=dict(
            usedefault=True,
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        out_prefix=dict(
            field="roptions.prefix",
            usedefault=True,
        ),
        paths=dict(),
        separation=dict(
            field="eoptions.sep",
        ),
        source=dict(
            copyfile=True,
            field="source",
            mandatory=True,
        ),
        target=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            field="ref",
            mandatory=True,
        ),
        tolerance=dict(
            field="eoptions.tol",
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
        write_interp=dict(
            field="roptions.interp",
        ),
        write_mask=dict(
            field="roptions.mask",
        ),
        write_wrap=dict(
            field="roptions.wrap",
        ),
    )
    inputs = Coregister.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Coregister_outputs():
    output_map = dict(
        coregistered_files=dict(),
        coregistered_source=dict(),
    )
    outputs = Coregister.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..tracking import Tractography


def test_Tractography_inputs():
    input_map = dict(
        act_file=dict(
            argstr="-act %s",
            extensions=None,
        ),
        algorithm=dict(
            argstr="-algorithm %s",
            usedefault=True,
        ),
        angle=dict(
            argstr="-angle %f",
        ),
        args=dict(
            argstr="%s",
        ),
        backtrack=dict(
            argstr="-backtrack",
        ),
        bval_scale=dict(
            argstr="-bvalue_scaling %s",
        ),
        crop_at_gmwmi=dict(
            argstr="-crop_at_gmwmi",
        ),
        cutoff=dict(
            argstr="-cutoff %f",
        ),
        cutoff_init=dict(
            argstr="-initcutoff %f",
        ),
        downsample=dict(
            argstr="-downsample %f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        grad_file=dict(
            argstr="-grad %s",
            extensions=None,
            xor=["grad_fsl"],
        ),
        grad_fsl=dict(
            argstr="-fslgrad %s %s",
            xor=["grad_file"],
        ),
        in_bval=dict(
            extensions=None,
        ),
        in_bvec=dict(
            argstr="-fslgrad %s %s",
            extensions=None,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        init_dir=dict(
            argstr="-initdirection %f,%f,%f",
        ),
        max_length=dict(
            argstr="-maxlength %f",
        ),
        max_seed_attempts=dict(
            argstr="-max_seed_attempts %d",
        ),
        max_tracks=dict(
            argstr="-maxnum %d",
        ),
        min_length=dict(
            argstr="-minlength %f",
        ),
        n_samples=dict(
            argstr="-samples %d",
            usedefault=True,
        ),
        n_tracks=dict(
            argstr="-number %d",
            max_ver="0.4",
        ),
        n_trials=dict(
            argstr="-trials %d",
        ),
        noprecompt=dict(
            argstr="-noprecomputed",
        ),
        nthreads=dict(
            argstr="-nthreads %d",
            nohash=True,
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-1,
            usedefault=True,
        ),
        out_seeds=dict(
            argstr="-output_seeds %s",
            extensions=None,
            usedefault=True,
        ),
        power=dict(
            argstr="-power %d",
        ),
        roi_excl=dict(
            argstr="-exclude %s",
        ),
        roi_incl=dict(
            argstr="-include %s",
        ),
        roi_mask=dict(
            argstr="-mask %s",
        ),
        seed_dynamic=dict(
            argstr="-seed_dynamic %s",
            extensions=None,
        ),
        seed_gmwmi=dict(
            argstr="-seed_gmwmi %s",
            extensions=None,
            requires=["act_file"],
        ),
        seed_grid_voxel=dict(
            argstr="-seed_grid_per_voxel %s %d",
            xor=["seed_image", "seed_rnd_voxel"],
        ),
        seed_image=dict(
            argstr="-seed_image %s",
            extensions=None,
        ),
        seed_rejection=dict(
            argstr="-seed_rejection %s",
            extensions=None,
        ),
        seed_rnd_voxel=dict(
            argstr="-seed_random_per_voxel %s %d",
            xor=["seed_image", "seed_grid_voxel"],
        ),
        seed_sphere=dict(
            argstr="-seed_sphere %f,%f,%f,%f",
        ),
        select=dict(
            argstr="-select %d",
            min_ver="3",
        ),
        sph_trait=dict(
            argstr="%f,%f,%f,%f",
        ),
        step_size=dict(
            argstr="-step %f",
        ),
        stop=dict(
            argstr="-stop",
        ),
        unidirectional=dict(
            argstr="-unidirectional",
        ),
        use_rk4=dict(
            argstr="-rk4",
        ),
    )
    inputs = Tractography.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Tractography_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
        out_seeds=dict(
            extensions=None,
        ),
    )
    outputs = Tractography.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import FUGUE


def test_FUGUE_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        asym_se_time=dict(
            argstr="--asym=%.10f",
        ),
        despike_2dfilter=dict(
            argstr="--despike",
        ),
        despike_threshold=dict(
            argstr="--despikethreshold=%s",
        ),
        dwell_time=dict(
            argstr="--dwell=%.10f",
        ),
        dwell_to_asym_ratio=dict(
            argstr="--dwelltoasym=%.10f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fmap_in_file=dict(
            argstr="--loadfmap=%s",
            extensions=None,
        ),
        fmap_out_file=dict(
            argstr="--savefmap=%s",
            extensions=None,
        ),
        forward_warping=dict(
            usedefault=True,
        ),
        fourier_order=dict(
            argstr="--fourier=%d",
        ),
        icorr=dict(
            argstr="--icorr",
            requires=["shift_in_file"],
        ),
        icorr_only=dict(
            argstr="--icorronly",
            requires=["unwarped_file"],
        ),
        in_file=dict(
            argstr="--in=%s",
            extensions=None,
        ),
        mask_file=dict(
            argstr="--mask=%s",
            extensions=None,
        ),
        median_2dfilter=dict(
            argstr="--median",
        ),
        no_extend=dict(
            argstr="--noextend",
        ),
        no_gap_fill=dict(
            argstr="--nofill",
        ),
        nokspace=dict(
            argstr="--nokspace",
        ),
        output_type=dict(),
        pava=dict(
            argstr="--pava",
        ),
        phase_conjugate=dict(
            argstr="--phaseconj",
        ),
        phasemap_in_file=dict(
            argstr="--phasemap=%s",
            extensions=None,
        ),
        poly_order=dict(
            argstr="--poly=%d",
        ),
        save_fmap=dict(
            xor=["save_unmasked_fmap"],
        ),
        save_shift=dict(
            xor=["save_unmasked_shift"],
        ),
        save_unmasked_fmap=dict(
            argstr="--unmaskfmap",
            xor=["save_fmap"],
        ),
        save_unmasked_shift=dict(
            argstr="--unmaskshift",
            xor=["save_shift"],
        ),
        shift_in_file=dict(
            argstr="--loadshift=%s",
            extensions=None,
        ),
        shift_out_file=dict(
            argstr="--saveshift=%s",
            extensions=None,
        ),
        smooth2d=dict(
            argstr="--smooth2=%.2f",
        ),
        smooth3d=dict(
            argstr="--smooth3=%.2f",
        ),
        unwarp_direction=dict(
            argstr="--unwarpdir=%s",
        ),
        unwarped_file=dict(
            argstr="--unwarp=%s",
            extensions=None,
            requires=["in_file"],
            xor=["warped_file"],
        ),
        warped_file=dict(
            argstr="--warp=%s",
            extensions=None,
            requires=["in_file"],
            xor=["unwarped_file"],
        ),
    )
    inputs = FUGUE.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FUGUE_outputs():
    output_map = dict(
        fmap_out_file=dict(
            extensions=None,
        ),
        shift_out_file=dict(
            extensions=None,
        ),
        unwarped_file=dict(
            extensions=None,
        ),
        warped_file=dict(
            extensions=None,
        ),
    )
    outputs = FUGUE.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import numpy as np
import nibabel as nb
from nipype.algorithms.stats import ActivationCount
import pytest


def test_ActivationCount(tmpdir):
    tmpdir.chdir()
    in_files = ["{:d}.nii".format(i) for i in range(3)]
    for fname in in_files:
        nb.Nifti1Image(np.random.normal(size=(5, 5, 5)), np.eye(4)).to_filename(fname)

    acm = ActivationCount(in_files=in_files, threshold=1.65)
    res = acm.run()
    diff = nb.load(res.outputs.out_file)
    pos = nb.load(res.outputs.acm_pos)
    neg = nb.load(res.outputs.acm_neg)
    assert np.allclose(diff.get_fdata(), pos.get_fdata() - neg.get_fdata())


@pytest.mark.parametrize(
    "threshold, above_thresh",
    [
        (1, 15.865),  # above one standard deviation (one side)
        (2, 2.275),  # above two standard deviations (one side)
        (3, 0.135),  # above three standard deviations (one side)
    ],
)
def test_ActivationCount_normaldistr(tmpdir, threshold, above_thresh):
    tmpdir.chdir()
    in_files = ["{:d}.nii".format(i) for i in range(3)]
    for fname in in_files:
        nb.Nifti1Image(np.random.normal(size=(100, 100, 100)), np.eye(4)).to_filename(
            fname
        )

    acm = ActivationCount(in_files=in_files, threshold=threshold)
    res = acm.run()
    pos = nb.load(res.outputs.acm_pos)
    neg = nb.load(res.outputs.acm_neg)
    assert np.isclose(
        pos.get_fdata().mean(), above_thresh * 1.0e-2, rtol=0.1, atol=1.0e-4
    )
    assert np.isclose(
        neg.get_fdata().mean(), above_thresh * 1.0e-2, rtol=0.1, atol=1.0e-4
    )



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..fiberprocess import fiberprocess


def test_fiberprocess_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        displacement_field=dict(
            argstr="--displacement_field %s",
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fiber_file=dict(
            argstr="--fiber_file %s",
            extensions=None,
        ),
        fiber_output=dict(
            argstr="--fiber_output %s",
            hash_files=False,
        ),
        fiber_radius=dict(
            argstr="--fiber_radius %f",
        ),
        h_field=dict(
            argstr="--h_field %s",
            extensions=None,
        ),
        index_space=dict(
            argstr="--index_space ",
        ),
        noDataChange=dict(
            argstr="--noDataChange ",
        ),
        no_warp=dict(
            argstr="--no_warp ",
        ),
        saveProperties=dict(
            argstr="--saveProperties ",
        ),
        tensor_volume=dict(
            argstr="--tensor_volume %s",
            extensions=None,
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
        voxel_label=dict(
            argstr="--voxel_label %d",
        ),
        voxelize=dict(
            argstr="--voxelize %s",
            hash_files=False,
        ),
        voxelize_count_fibers=dict(
            argstr="--voxelize_count_fibers ",
        ),
    )
    inputs = fiberprocess.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_fiberprocess_outputs():
    output_map = dict(
        fiber_output=dict(
            extensions=None,
        ),
        voxelize=dict(
            extensions=None,
        ),
    )
    outputs = fiberprocess.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import FEATModel


def test_FEATModel_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ev_files=dict(
            argstr="%s",
            copyfile=False,
            mandatory=True,
            position=1,
        ),
        fsf_file=dict(
            argstr="%s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=0,
        ),
        output_type=dict(),
    )
    inputs = FEATModel.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FEATModel_outputs():
    output_map = dict(
        con_file=dict(
            extensions=None,
        ),
        design_cov=dict(
            extensions=None,
        ),
        design_file=dict(
            extensions=None,
        ),
        design_image=dict(
            extensions=None,
        ),
        fcon_file=dict(
            extensions=None,
        ),
    )
    outputs = FEATModel.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..tracking import StreamlineTrack


def test_StreamlineTrack_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        cutoff_value=dict(
            argstr="-cutoff %s",
            units="NA",
        ),
        desired_number_of_tracks=dict(
            argstr="-number %d",
        ),
        do_not_precompute=dict(
            argstr="-noprecomputed",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        exclude_file=dict(
            argstr="-exclude %s",
            extensions=None,
            xor=["exclude_file", "exclude_spec"],
        ),
        exclude_spec=dict(
            argstr="-exclude %s",
            position=2,
            sep=",",
            units="mm",
            xor=["exclude_file", "exclude_spec"],
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        include_file=dict(
            argstr="-include %s",
            extensions=None,
            xor=["include_file", "include_spec"],
        ),
        include_spec=dict(
            argstr="-include %s",
            position=2,
            sep=",",
            units="mm",
            xor=["include_file", "include_spec"],
        ),
        initial_cutoff_value=dict(
            argstr="-initcutoff %s",
            units="NA",
        ),
        initial_direction=dict(
            argstr="-initdirection %s",
            units="voxels",
        ),
        inputmodel=dict(
            argstr="%s",
            position=-3,
            usedefault=True,
        ),
        mask_file=dict(
            argstr="-mask %s",
            extensions=None,
            xor=["mask_file", "mask_spec"],
        ),
        mask_spec=dict(
            argstr="-mask %s",
            position=2,
            sep=",",
            units="mm",
            xor=["mask_file", "mask_spec"],
        ),
        maximum_number_of_tracks=dict(
            argstr="-maxnum %d",
        ),
        maximum_tract_length=dict(
            argstr="-length %s",
            units="mm",
        ),
        minimum_radius_of_curvature=dict(
            argstr="-curvature %s",
            units="mm",
        ),
        minimum_tract_length=dict(
            argstr="-minlength %s",
            units="mm",
        ),
        no_mask_interpolation=dict(
            argstr="-nomaskinterp",
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_tracked.tck",
            output_name="tracked",
            position=-1,
        ),
        seed_file=dict(
            argstr="-seed %s",
            extensions=None,
            xor=["seed_file", "seed_spec"],
        ),
        seed_spec=dict(
            argstr="-seed %s",
            position=2,
            sep=",",
            units="mm",
            xor=["seed_file", "seed_spec"],
        ),
        step_size=dict(
            argstr="-step %s",
            units="mm",
        ),
        stop=dict(
            argstr="-stop",
        ),
        unidirectional=dict(
            argstr="-unidirectional",
        ),
    )
    inputs = StreamlineTrack.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_StreamlineTrack_outputs():
    output_map = dict(
        tracked=dict(
            extensions=None,
        ),
    )
    outputs = StreamlineTrack.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import Localstat


def test_Localstat_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        automask=dict(
            argstr="-automask",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        grid_rmode=dict(
            argstr="-grid_rmode %s",
            requires=["reduce_restore_grid"],
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        mask_file=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        neighborhood=dict(
            argstr="-nbhd '%s(%s)'",
            mandatory=True,
        ),
        nonmask=dict(
            argstr="-use_nonmask",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            keep_extension=True,
            name_source="in_file",
            name_template="%s_localstat",
            position=0,
        ),
        outputtype=dict(),
        overwrite=dict(
            argstr="-overwrite",
        ),
        quiet=dict(
            argstr="-quiet",
        ),
        reduce_grid=dict(
            argstr="-reduce_grid %s",
            xor=["reduce_restore_grid", "reduce_max_vox"],
        ),
        reduce_max_vox=dict(
            argstr="-reduce_max_vox %s",
            xor=["reduce_restore_grid", "reduce_grid"],
        ),
        reduce_restore_grid=dict(
            argstr="-reduce_restore_grid %s",
            xor=["reduce_max_vox", "reduce_grid"],
        ),
        stat=dict(
            argstr="-stat %s...",
            mandatory=True,
        ),
    )
    inputs = Localstat.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Localstat_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = Localstat.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..diffusion import dtiprocess


def test_dtiprocess_inputs():
    input_map = dict(
        DTI_double=dict(
            argstr="--DTI_double ",
        ),
        RD_output=dict(
            argstr="--RD_output %s",
            hash_files=False,
        ),
        affineitk_file=dict(
            argstr="--affineitk_file %s",
            extensions=None,
        ),
        args=dict(
            argstr="%s",
        ),
        color_fa_output=dict(
            argstr="--color_fa_output %s",
            hash_files=False,
        ),
        correction=dict(
            argstr="--correction %s",
        ),
        deformation_output=dict(
            argstr="--deformation_output %s",
            hash_files=False,
        ),
        dof_file=dict(
            argstr="--dof_file %s",
            extensions=None,
        ),
        dti_image=dict(
            argstr="--dti_image %s",
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fa_gradient_output=dict(
            argstr="--fa_gradient_output %s",
            hash_files=False,
        ),
        fa_gradmag_output=dict(
            argstr="--fa_gradmag_output %s",
            hash_files=False,
        ),
        fa_output=dict(
            argstr="--fa_output %s",
            hash_files=False,
        ),
        forward=dict(
            argstr="--forward %s",
            extensions=None,
        ),
        frobenius_norm_output=dict(
            argstr="--frobenius_norm_output %s",
            hash_files=False,
        ),
        hField=dict(
            argstr="--hField ",
        ),
        interpolation=dict(
            argstr="--interpolation %s",
        ),
        lambda1_output=dict(
            argstr="--lambda1_output %s",
            hash_files=False,
        ),
        lambda2_output=dict(
            argstr="--lambda2_output %s",
            hash_files=False,
        ),
        lambda3_output=dict(
            argstr="--lambda3_output %s",
            hash_files=False,
        ),
        mask=dict(
            argstr="--mask %s",
            extensions=None,
        ),
        md_output=dict(
            argstr="--md_output %s",
            hash_files=False,
        ),
        negative_eigenvector_output=dict(
            argstr="--negative_eigenvector_output %s",
            hash_files=False,
        ),
        newdof_file=dict(
            argstr="--newdof_file %s",
            extensions=None,
        ),
        outmask=dict(
            argstr="--outmask %s",
            hash_files=False,
        ),
        principal_eigenvector_output=dict(
            argstr="--principal_eigenvector_output %s",
            hash_files=False,
        ),
        reorientation=dict(
            argstr="--reorientation %s",
        ),
        rot_output=dict(
            argstr="--rot_output %s",
            hash_files=False,
        ),
        scalar_float=dict(
            argstr="--scalar_float ",
        ),
        sigma=dict(
            argstr="--sigma %f",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
    )
    inputs = dtiprocess.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_dtiprocess_outputs():
    output_map = dict(
        RD_output=dict(
            extensions=None,
        ),
        color_fa_output=dict(
            extensions=None,
        ),
        deformation_output=dict(
            extensions=None,
        ),
        fa_gradient_output=dict(
            extensions=None,
        ),
        fa_gradmag_output=dict(
            extensions=None,
        ),
        fa_output=dict(
            extensions=None,
        ),
        frobenius_norm_output=dict(
            extensions=None,
        ),
        lambda1_output=dict(
            extensions=None,
        ),
        lambda2_output=dict(
            extensions=None,
        ),
        lambda3_output=dict(
            extensions=None,
        ),
        md_output=dict(
            extensions=None,
        ),
        negative_eigenvector_output=dict(
            extensions=None,
        ),
        outmask=dict(
            extensions=None,
        ),
        principal_eigenvector_output=dict(
            extensions=None,
        ),
        rot_output=dict(
            extensions=None,
        ),
    )
    outputs = dtiprocess.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..odf import ODFTracker


def test_ODFTracker_inputs():
    input_map = dict(
        ODF=dict(
            extensions=None,
            mandatory=True,
        ),
        angle_threshold=dict(
            argstr="-at %f",
        ),
        args=dict(
            argstr="%s",
        ),
        disc=dict(
            argstr="-disc",
        ),
        dsi=dict(
            argstr="-dsi",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        image_orientation_vectors=dict(
            argstr="-iop %f",
        ),
        input_data_prefix=dict(
            argstr="%s",
            position=0,
            usedefault=True,
        ),
        input_output_type=dict(
            argstr="-it %s",
            usedefault=True,
        ),
        invert_x=dict(
            argstr="-ix",
        ),
        invert_y=dict(
            argstr="-iy",
        ),
        invert_z=dict(
            argstr="-iz",
        ),
        limit=dict(
            argstr="-limit %d",
        ),
        mask1_file=dict(
            argstr="-m %s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        mask1_threshold=dict(
            position=3,
        ),
        mask2_file=dict(
            argstr="-m2 %s",
            extensions=None,
            position=4,
        ),
        mask2_threshold=dict(
            position=5,
        ),
        max=dict(
            extensions=None,
            mandatory=True,
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            position=1,
            usedefault=True,
        ),
        random_seed=dict(
            argstr="-rseed %s",
        ),
        runge_kutta2=dict(
            argstr="-rk2",
        ),
        slice_order=dict(
            argstr="-sorder %d",
        ),
        step_length=dict(
            argstr="-l %f",
        ),
        swap_xy=dict(
            argstr="-sxy",
        ),
        swap_yz=dict(
            argstr="-syz",
        ),
        swap_zx=dict(
            argstr="-szx",
        ),
        voxel_order=dict(
            argstr="-vorder %s",
        ),
    )
    inputs = ODFTracker.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ODFTracker_outputs():
    output_map = dict(
        track_file=dict(
            extensions=None,
        ),
    )
    outputs = ODFTracker.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..patchmatch import PatchMatch


def test_PatchMatch_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        cs_size=dict(
            argstr="-cs %i",
        ),
        database_file=dict(
            argstr="-db %s",
            extensions=None,
            mandatory=True,
            position=3,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        it_num=dict(
            argstr="-it %i",
        ),
        mask_file=dict(
            argstr="-m %s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        match_num=dict(
            argstr="-match %i",
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_pm.nii.gz",
            position=4,
        ),
        patch_size=dict(
            argstr="-size %i",
        ),
        pm_num=dict(
            argstr="-pm %i",
        ),
    )
    inputs = PatchMatch.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_PatchMatch_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = PatchMatch.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
from shutil import rmtree

import pytest

from nipype.utils.misc import container_to_string, str2bool, flatten, unflatten


def test_cont_to_str():
    # list
    x = ["a", "b"]
    assert container_to_string(x) == "a b"
    # tuple
    x = tuple(x)
    assert container_to_string(x) == "a b"
    # set
    x = set(x)
    y = container_to_string(x)
    assert (y == "a b") or (y == "b a")
    # dict
    x = dict(a="a", b="b")
    y = container_to_string(x)
    assert (y == "a b") or (y == "b a")
    # string
    assert container_to_string("foobar") == "foobar"
    # int.  Integers are not the main intent of this function, but see
    # no reason why they shouldn't work.
    assert container_to_string(123) == "123"


@pytest.mark.parametrize(
    "string, expected",
    [
        ("yes", True),
        ("true", True),
        ("t", True),
        ("1", True),
        ("no", False),
        ("false", False),
        ("n", False),
        ("f", False),
        ("0", False),
    ],
)
def test_str2bool(string, expected):
    assert str2bool(string) == expected


def test_flatten():
    in_list = [[1, 2, 3], [4], [[5, 6], 7], 8]

    flat = flatten(in_list)
    assert flat == [1, 2, 3, 4, 5, 6, 7, 8]

    back = unflatten(flat, in_list)
    assert in_list == back

    new_list = [2, 3, 4, 5, 6, 7, 8, 9]
    back = unflatten(new_list, in_list)
    assert back == [[2, 3, 4], [5], [[6, 7], 8], 9]

    flat = flatten([])
    assert flat == []

    back = unflatten([], [])
    assert back == []


def test_rgetcwd(monkeypatch, tmpdir):
    from ..misc import rgetcwd

    oldpath = tmpdir.strpath
    tmpdir.mkdir("sub").chdir()
    newpath = os.getcwd()

    # Path still there
    assert rgetcwd() == newpath

    # Remove path
    rmtree(newpath, ignore_errors=True)
    with pytest.raises(OSError):
        os.getcwd()

    monkeypatch.setenv("PWD", oldpath)
    assert rgetcwd(error=False) == oldpath

    # Test when error should be raised
    with pytest.raises(OSError):
        rgetcwd()

    # Deleted env variable
    monkeypatch.delenv("PWD")
    with pytest.raises(OSError):
        rgetcwd(error=False)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..petpvc import PETPVC


def test_PETPVC_inputs():
    input_map = dict(
        alpha=dict(
            argstr="-a %.4f",
            usedefault=True,
        ),
        args=dict(
            argstr="%s",
        ),
        debug=dict(
            argstr="-d",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fwhm_x=dict(
            argstr="-x %.4f",
            mandatory=True,
        ),
        fwhm_y=dict(
            argstr="-y %.4f",
            mandatory=True,
        ),
        fwhm_z=dict(
            argstr="-z %.4f",
            mandatory=True,
        ),
        in_file=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        mask_file=dict(
            argstr="-m %s",
            extensions=None,
            mandatory=True,
        ),
        n_deconv=dict(
            argstr="-k %d",
            usedefault=True,
        ),
        n_iter=dict(
            argstr="-n %d",
            usedefault=True,
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
        pvc=dict(
            argstr="-p %s",
            mandatory=True,
        ),
        stop_crit=dict(
            argstr="-s %.4f",
            usedefault=True,
        ),
    )
    inputs = PETPVC.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_PETPVC_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = PETPVC.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os

import nibabel as nb
import numpy as np

import pytest
from ...testing import utils
from ..confounds import CompCor, TCompCor, ACompCor


class TestCompCor:
    """Note: Tests currently do a poor job of testing functionality"""

    filenames = {
        "functionalnii": "compcorfunc.nii",
        "masknii": "compcormask.nii",
        "masknii2": "compcormask2.nii",
        "components_file": None,
    }

    @pytest.fixture(autouse=True)
    def setup_class(self, tmpdir):
        # setup
        tmpdir.chdir()
        noise = np.fromfunction(self.fake_noise_fun, self.fake_data.shape)
        self.realigned_file = utils.save_toy_nii(
            self.fake_data + noise, self.filenames["functionalnii"]
        )
        mask = np.ones(self.fake_data.shape[:3])
        mask[0, 0, 0] = 0
        mask[0, 0, 1] = 0
        mask1 = utils.save_toy_nii(mask, self.filenames["masknii"])

        other_mask = np.ones(self.fake_data.shape[:3])
        other_mask[0, 1, 0] = 0
        other_mask[1, 1, 0] = 0
        mask2 = utils.save_toy_nii(other_mask, self.filenames["masknii2"])

        self.mask_files = [mask1, mask2]

    def test_compcor(self):
        expected_components = [
            ["-0.1989607212", "-0.5753813646"],
            ["0.5692369697", "0.5674945949"],
            ["-0.6662573243", "0.4675843432"],
            ["0.4206466244", "-0.3361270124"],
            ["-0.1246655485", "-0.1235705610"],
        ]

        self.run_cc(
            CompCor(
                num_components=6,
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                mask_index=0,
            ),
            expected_components,
        )

        self.run_cc(
            ACompCor(
                num_components=6,
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                mask_index=0,
                components_file="acc_components_file",
            ),
            expected_components,
            "aCompCor",
        )

    def test_compcor_variance_threshold_and_metadata(self):
        expected_components = [
            ["-0.2027150345", "-0.4954813834"],
            ["0.2565929051", "0.7866217875"],
            ["-0.3550986008", "-0.0089784905"],
            ["0.7512786244", "-0.3599828482"],
            ["-0.4500578942", "0.0778209345"],
        ]
        expected_metadata = {
            "component": "CompCor00",
            "mask": "mask",
            "singular_value": "4.0720553036",
            "variance_explained": "0.5527211465",
            "cumulative_variance_explained": "0.5527211465",
            "retained": "True",
        }
        ccinterface = CompCor(
            variance_threshold=0.7,
            realigned_file=self.realigned_file,
            mask_files=self.mask_files,
            mask_names=["mask"],
            mask_index=1,
            save_metadata=True,
        )
        self.run_cc(
            ccinterface=ccinterface,
            expected_components=expected_components,
            expected_n_components=2,
            expected_metadata=expected_metadata,
        )

    def test_tcompcor(self):
        ccinterface = TCompCor(
            num_components=6,
            realigned_file=self.realigned_file,
            percentile_threshold=0.75,
        )
        self.run_cc(
            ccinterface,
            [
                ["-0.1114536190", "-0.4632908609"],
                ["0.4566907310", "0.6983205193"],
                ["-0.7132557407", "0.1340170559"],
                ["0.5022537643", "-0.5098322262"],
                ["-0.1342351356", "0.1407855119"],
            ],
            "tCompCor",
        )

    def test_tcompcor_no_percentile(self):
        ccinterface = TCompCor(num_components=6, realigned_file=self.realigned_file)
        ccinterface.run()

        mask = nb.load("mask_000.nii.gz").dataobj
        num_nonmasked_voxels = np.count_nonzero(mask)
        assert num_nonmasked_voxels == 1

    def test_compcor_no_regress_poly(self):
        self.run_cc(
            CompCor(
                num_components=6,
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                mask_index=0,
                pre_filter=False,
            ),
            [
                ["0.4451946442", "-0.7683311482"],
                ["-0.4285129505", "-0.0926034137"],
                ["0.5721540256", "0.5608764842"],
                ["-0.5367548139", "0.0059943226"],
                ["-0.0520809054", "0.2940637551"],
            ],
        )

    def test_tcompcor_asymmetric_dim(self):
        asymmetric_shape = (2, 3, 4, 5)
        asymmetric_data = utils.save_toy_nii(
            np.zeros(asymmetric_shape), "asymmetric.nii"
        )

        TCompCor(realigned_file=asymmetric_data).run()
        assert nb.load("mask_000.nii.gz").shape == asymmetric_shape[:3]

    def test_compcor_bad_input_shapes(self):
        # dim 0 is < dim 0 of self.mask_files (2)
        shape_less_than = (1, 2, 2, 5)
        # dim 0 is > dim 0 of self.mask_files (2)
        shape_more_than = (3, 3, 3, 5)

        for data_shape in (shape_less_than, shape_more_than):
            data_file = utils.save_toy_nii(np.zeros(data_shape), "temp.nii")
            interface = CompCor(realigned_file=data_file, mask_files=self.mask_files[0])
            with pytest.raises(ValueError):
                interface.run()  # Dimension mismatch

    def test_tcompcor_bad_input_dim(self):
        bad_dims = (2, 2, 2)
        data_file = utils.save_toy_nii(np.zeros(bad_dims), "temp.nii")
        interface = TCompCor(realigned_file=data_file)
        with pytest.raises(ValueError):
            interface.run()  # Not a 4D file

    def test_tcompcor_merge_intersect_masks(self):
        for method in ["union", "intersect"]:
            TCompCor(
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                merge_method=method,
            ).run()
            if method == "union":
                assert np.array_equal(
                    nb.load("mask_000.nii.gz").dataobj,
                    ([[[0, 0], [0, 0]], [[0, 0], [1, 0]]]),
                )
            if method == "intersect":
                assert np.array_equal(
                    nb.load("mask_000.nii.gz").dataobj,
                    ([[[0, 0], [0, 0]], [[0, 1], [0, 0]]]),
                )

    def test_tcompcor_index_mask(self):
        TCompCor(
            realigned_file=self.realigned_file, mask_files=self.mask_files, mask_index=1
        ).run()
        assert np.array_equal(
            nb.load("mask_000.nii.gz").dataobj, ([[[0, 0], [0, 0]], [[0, 1], [0, 0]]])
        )

    def test_tcompcor_multi_mask_no_index(self):
        interface = TCompCor(
            realigned_file=self.realigned_file, mask_files=self.mask_files
        )
        with pytest.raises(ValueError):
            interface.run()  # more than one mask file

    def run_cc(
        self,
        ccinterface,
        expected_components,
        expected_header="CompCor",
        expected_n_components=None,
        expected_metadata=None,
    ):
        # run
        ccresult = ccinterface.run()

        # assert
        expected_file = ccinterface._list_outputs()["components_file"]
        assert ccresult.outputs.components_file == expected_file
        assert os.path.exists(expected_file)
        assert os.path.getsize(expected_file) > 0

        with open(ccresult.outputs.components_file, "r") as components_file:
            if expected_n_components is None:
                expected_n_components = min(
                    ccinterface.inputs.num_components, self.fake_data.shape[3]
                )

            components_data = [line.rstrip().split("\t") for line in components_file]

            # the first item will be '#', we can throw it out
            header = components_data.pop(0)
            expected_header = [
                expected_header + "{:02d}".format(i)
                for i in range(expected_n_components)
            ]
            for i, heading in enumerate(header):
                assert expected_header[i] in heading

            num_got_timepoints = len(components_data)
            assert num_got_timepoints == self.fake_data.shape[3]
            for index, timepoint in enumerate(components_data):
                assert len(timepoint) == expected_n_components
                assert timepoint[:2] == expected_components[index]

        if ccinterface.inputs.save_metadata:
            expected_metadata_file = ccinterface._list_outputs()["metadata_file"]
            assert ccresult.outputs.metadata_file == expected_metadata_file
            assert os.path.exists(expected_metadata_file)
            assert os.path.getsize(expected_metadata_file) > 0

            with open(ccresult.outputs.metadata_file, "r") as metadata_file:
                components_metadata = [
                    line.rstrip().split("\t") for line in metadata_file
                ]
                components_metadata = {
                    i: j for i, j in zip(components_metadata[0], components_metadata[1])
                }
                assert components_metadata == expected_metadata

        return ccresult

    @staticmethod
    def fake_noise_fun(i, j, l, m):
        return m * i + l - j

    fake_data = np.array(
        [
            [[[8, 5, 3, 8, 0], [6, 7, 4, 7, 1]], [[7, 9, 1, 6, 5], [0, 7, 4, 7, 7]]],
            [[[2, 4, 5, 7, 0], [1, 7, 0, 5, 4]], [[7, 3, 9, 0, 4], [9, 4, 1, 5, 0]]],
        ]
    )
